
def Add(shapeA, shapeB, axis = None, broadcast = None):
    #do not deal
    return shapeA

def AveragePool(shape, auto_pad = None, kernelShape = None, pads = None, strides = None):
    #I don't want to deal with auto_pad
   
    if kernelShape is None:
        kernelShape = [2 for _ in range(2)]

    dim = len(kernelShape)

    if pads is None:
        pads = [1 for _ in range(dim * 2)]
    if strides is None:
        strides = [1 for _ in range(dim)]
 
    retShape = shape[:-dim]
    dimIdx = 0
    for dimSize in shape[-dim:]:
        padUpper = pads[dimIdx * 2]
        padLower = pads[dimIdx * 2 + 1]
        stride = strides[dimIdx]
        kernelDimSize = kernelShape[dimIdx]
        retShape.append((dimSize + padUpper + padLower - kernelDimSize) // stride + 1)
        dimIdx = dimIdx + 1

    return retShape


def BatchNormalization(shape, scale = None, B = None, mean = None, var = None):
    return shape

def Concat(shapeList, axis):
    newDimSize = sum([x[axis] for x in shapeList])
    newShape = shapeList[0]  
    newShape[axis] = newDimSize
    return newShape

def Conv(shapeX, shapeW, auto_pad = None, dilations = None, group = None, kernel_shape = None, pads = None, strides = None):
    #Don't support auto_pad current!
    #                             2018-02-28
    #if group is None:
    #    group = 1
    #  group is not support yet too.
    kernelDim = len(shapeX) - 2
    if kernel_shape is None:
        kernel_shape = shapeW[2:] #[[1 for _ in range(kernelDimSize)] for _ in range(kernelDimSize)]
    if pads is None:
        [0 for _ in range(kernelDim * 2)]
    if strides is None:
        [1 for _ in range(kernelDim)]
    if pads is None:
        pads = [0 for _ in range(kernelDim * 2)]
    if strides is None:
        strides = [1 for _ in range(kernelDim)]
    if dilations is None:
        dilations = [1 for _ in range(kernelDim)]
 
    retShape = [shapeX[0], shapeW[0]]  
    dimIdx = 0
    for dimSize in shapeX[2:]:
        padUpper = pads[dimIdx * 2]
        padLower = pads[dimIdx * 2 + 1]
        stride = strides[dimIdx]
        dilation = dilations[dimIdx]
        kernelDimSize = (kernel_shape[dimIdx] - 1) // 2 * dilation * 2 + 1
        retShape.append((dimSize + padUpper + padLower - kernelDimSize) // stride + 1)
        dimIdx = dimIdx + 1
    return retShape

def GlobalAveragePool(shapeX):
    return shapeX[:2] + [1, 1]

def MaxPool(shape, auto_pad = None, kernelShape = None, pads = None, strides = None):
    return AveragePool(shape, auto_pad, kernelShape, pads, strides)

def Mul(shapeX, shapeW, axis = None, broadcast = None):
    return shapeX

def Relu(shape):
    return shape

def FC(shapeX, shapeW, shapeB = None, axis = None, axis_w = None):
    if axis is None:
        axis = 1
    if axis_w is None:
        axis_w = 1
    return [shapeX[0], shapeW[1]]

def Flatten(shapeT, axis = None):
    if axis is None:
        axis = 1

    firstDim = 1
    secondDim = 1
    for i in range(len(shapeT)):
        if i < axis:
            firstDim *= shapeT[i]
        else:
            secondDim *= shapeT[i]
    
    if (axis > 0):
        return [firstDim, secondDim]
    else:
        return [secondDim]

inference_shape = {
    'Add' : Add,
    'AveragePool' : AveragePool,
    'BatchNormalization' : BatchNormalization,
    'Concat' : Concat,
    'Conv' : Conv,
    'GlobalAveragePool' : GlobalAveragePool,
    'MaxPool' : MaxPool,
    'Mul' : Mul,
    'Relu' : Relu,
    'FC' : FC,
    'Flatten' : Flatten
}

def testByLeNet(image_shape):
    print('\nLeNet output shape test:')
    print('input_image_shape is : ', image_shape)
    convLay1 = [5, 5]
    WLay1 = [6, -1, 5, 5]
    outputLay1 = inference_shape['Conv'](image_shape, WLay1, kernel_shape = convLay1)
    print('1st Lay output shape is : ', outputLay1)  

    poolLay2 = [2, 2]
    stridesLay2 = [2, 2]
    outputLay2 = inference_shape['AveragePool'](outputLay1, strides = stridesLay2)
    print('2nd Lay output shape is : ', outputLay2)

    convLay3 = [5, 5]
    WLay3 = [16, -1, 5, 5]
    outputLay3 = inference_shape['Conv'](outputLay2, WLay3, kernel_shape = convLay3)
    print('3rd Lay output shape is : ', outputLay3)

    poolLay4 = [2, 2]
    stridesLay4 = [2, 2]
    outputLay4 = inference_shape['AveragePool'](outputLay3, strides = stridesLay4)
    print('4th Lay output shape is : ', outputLay4)

    convLay5 = [5, 5]
    WLay5 = [120, -1, 5, 5]
    outputLay5 = inference_shape['Conv'](outputLay4, WLay5)
    print('5th Lay output shape is : ', outputLay5)
    
    outputLay5Flatten = inference_shape['Flatten'](outputLay5)
    WLay6 = [-1, 84]
    outputLay6 = inference_shape['FC'](outputLay5Flatten, WLay6)
    print('6th Lay output shape is : ', outputLay6)

    WLay7 = [-1, 10]
    outputLay7 = inference_shape['FC'](outputLay6, WLay7)
    print('7th Lay output shape is : ', outputLay7)
    return outputLay7


if __name__ == '__main__':

    shape = [1, 9, 9, 9]

    print('input shape is  : ', shape)
    print('output shape is : ', AveragePool(shape, pads=[1,1,1,1], strides=[2,2]))
    print(inference_shape['AveragePool'](shape, pads=[1,1,1,1], kernelShape=[2,2], strides=[2,2]));

    print('input shape is  : ', shape)
    print('output shape is : ', AveragePool(shape, pads=[0,0,0,0], kernelShape=[3,3], strides=[3,3]))

    shape = [3, 9, 9]
    print('input shape is  : ', shape)
    print('output shape is : ', AveragePool(shape))


    x = [1, 1, 5, 5]
    W = [1, 1, 3, 3]
    print('input shapeX is :', x, 'input shapeW is :', W)
    print('output shape is :', Conv(x, W), "without pads")

    W = [2, 1, 3, 3]
    print('input shapeX is :', x, 'input shapeW is :', W)
    print('output shape is :', Conv(x, W, pads=[1,1,1,1]), 'pads is [1, 1, 1, 1]')


    shape1 = [1, 1, 3, 3]
    shape2 = [1, 3, 3, 3]
    shape3 = [1, 5, 3, 3]
    print('output shape is :', Concat([shape1, shape2, shape3], 1))

    shape = [5, 5, 5, 5]
    print("output shape is :", GlobalAveragePool(shape))
    
    print('LeNet-5 output shape is : ', testByLeNet(image_shape = [-1, 1, 32, 32]))
    print('LeNet-5 output shape is : ', testByLeNet(image_shape = [5, 1, 32, 32]))






import tensorflow as tf
from keras import backend as K
from PIL import Image

def letterbox_image(image, size):
    '''resize image with unchanged aspect ratio using padding'''
    image_w, image_h = image.size
    w, h = size
    new_w = int(image_w * min(w/image_w, h/image_h))
    new_h = int(image_h * min(w/image_w, h/image_h))
    resized_image = image.resize((new_w,new_h), Image.BICUBIC)

    boxed_image = Image.new('RGB', size, (128,128,128))
    boxed_image.paste(resized_image, ((w-new_w)//2,(h-new_h)//2))
    return boxed_image

def yolo_head(feats, anchors, num_classes, input_shape):
    """Convert final layer features to bounding box parameters."""
    num_anchors = len(anchors)
    # Reshape to batch, height, width, num_anchors, box_params.
    anchors_tensor = K.reshape(K.constant(anchors), [1, 1, 1, num_anchors, 2])

    conv_dims = K.shape(feats)[1:3]
    conv_height_index = K.arange(0, stop=conv_dims[1])
    conv_width_index = K.arange(0, stop=conv_dims[0])
    conv_height_index = K.tile(conv_height_index, [conv_dims[0]])

    conv_width_index = K.tile(
        K.expand_dims(conv_width_index, 0), [conv_dims[1], 1])
    conv_width_index = K.flatten(K.transpose(conv_width_index))
    conv_index = K.transpose(K.stack([conv_height_index, conv_width_index]))
    conv_index = K.reshape(conv_index, [1, conv_dims[0], conv_dims[1], 1, 2])
    conv_index = K.cast(conv_index, K.dtype(feats))

    feats = K.reshape(
        feats, [-1, conv_dims[0], conv_dims[1], num_anchors, num_classes + 5])
    conv_dims = K.cast(conv_dims[::-1], K.dtype(feats))

    box_xy = K.sigmoid(feats[..., :2])
    box_wh = K.exp(feats[..., 2:4])
    box_confidence = K.sigmoid(feats[..., 4:5])
    box_class_probs = K.sigmoid(feats[..., 5:])

    # Adjust preditions to each spatial grid point and anchor size.
    # Note: YOLO iterates over height index before width index.
    # TODO: It works with +1, don't know why.
    box_xy = (box_xy + conv_index + 1) / conv_dims
    box_wh = box_wh * anchors_tensor / K.cast(input_shape[::-1], K.dtype(box_wh))

    return box_xy, box_wh, box_confidence, box_class_probs


def yolo_correct_boxes(box_xy, box_wh, input_shape, image_shape):
    '''Get corrected boxes'''
    box_yx = box_xy[..., ::-1]
    box_hw = box_wh[..., ::-1]
    input_shape = K.cast(input_shape, K.dtype(box_yx))
    image_shape = K.cast(image_shape, K.dtype(box_yx))
    new_shape = K.round(image_shape * K.min(input_shape/image_shape))
    offset = (input_shape-new_shape)/2./input_shape
    scale = input_shape/new_shape
    box_yx = (box_yx - offset) * scale
    box_hw *= scale

    box_mins = box_yx - (box_hw / 2.)
    box_maxes = box_yx + (box_hw / 2.)
    boxes =  K.concatenate([
        box_mins[..., 0:1],  # y_min
        box_mins[..., 1:2],  # x_min
        box_maxes[..., 0:1],  # y_max
        box_maxes[..., 1:2]  # x_max
    ])

    # Scale boxes back to original image shape.
    boxes *= K.concatenate([image_shape, image_shape])
    return boxes

def yolo_boxes_and_scores(feats, anchors, num_classes, input_shape, image_shape):
    '''Process Conv layer output'''
    # print("feats,anchors, num_classes, input_shape", feats, anchors, num_classes, input_shape)
    box_xy, box_wh, box_confidence, box_class_probs = yolo_head(feats,
        anchors, num_classes, input_shape)
    # print(box_xy, box_wh, box_confidence, box_class_probs)
    boxes = yolo_correct_boxes(box_xy, box_wh, input_shape, image_shape)
    boxes = K.reshape(boxes, [-1, 4])
    box_scores = box_confidence * box_class_probs
    box_scores = K.reshape(box_scores, [-1, num_classes])
    return boxes, box_scores


def yolo_eval(yolo_outputs,
        anchors,
        num_classes,
        image_shape,
        max_boxes=20,
        score_threshold=.6,
        iou_threshold=.5):
    """Evaluate YOLO model on given input and return filtered boxes."""
    # yolo_outputs order 13,26,52

    input_shape = K.shape(yolo_outputs[0])[1:3] * 32

    for i in range(0,3):
        _boxes, _box_scores = yolo_boxes_and_scores(yolo_outputs[i],
            anchors[6-3*i:9-3*i], num_classes, input_shape, image_shape)
        if i==0:
            boxes, box_scores = _boxes, _box_scores
        else:
            boxes = K.concatenate([boxes,_boxes], axis=0)
            box_scores = K.concatenate([box_scores,_box_scores], axis=0)

    mask = box_scores >= score_threshold
    max_boxes_tensor = K.constant(max_boxes, dtype='int32')
    for i in range(num_classes):
        # TODO: use keras backend instead of tf.
        class_boxes = tf.boolean_mask(boxes, mask[:, i])
        class_box_scores = tf.boolean_mask(box_scores[:, i], mask[:, i])
        nms_index = tf.image.non_max_suppression(
            class_boxes, class_box_scores, max_boxes_tensor, iou_threshold=iou_threshold)
        class_boxes = K.gather(class_boxes, nms_index)
        class_box_scores = K.gather(class_box_scores, nms_index)
        classes = K.ones_like(class_box_scores, 'int32') * i
        if i==0:
            boxes_, scores_, classes_ = class_boxes, class_box_scores, classes
        else:
            boxes_ = K.concatenate([boxes_,class_boxes], axis=0)
            scores_ = K.concatenate([scores_,class_box_scores], axis=0)
            classes_ = K.concatenate([classes_,classes], axis=0)
    return boxes_, scores_, classes_


import numpy as np
import tensorflow
from tensorflow.python.framework import tensor_util
from tensorflow.core.framework import attr_value_pb2
from mmdnn.conversion.tensorflow.tensorflow_graph import TensorflowGraph
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.common.DataStructure.parser import Parser
from distutils.version import LooseVersion


class TensorflowParser2(Parser):

    # skip_prefix = [
    #     "^",
    #     "train_op",
    #     "save",
    #     "gradients",
    #     "init",
    #     "global_step",
    #     "distort_image",
    #     "Adagrad",
    # ]
    skip_prefix = [
    ]

    skip_scope = [
        "random_uniform",
        "Initializer",
        "optimizer",
        "weight_loss",
        "parallel_read",
        "case"
    ]

    skip_type = set([
        "L2Loss",
        "VariableV2",
        "Const",
        "Assign",
        "RandomUniform",
        "FIFOQueueV2",
        "Assert",
        "Unpack",
        "NextIteration",
        "TensorArrayV3",
        "Range",
        "TensorArrayScatterV3",
        "TensorArrayReadV3",
        "TensorArrayWriteV3",
        # "Switch"
        "Dequantize",
        # "RequantizationRange",
        # "Requantize",
        "ExpandDims",
        # "Identity",
        # "Mean",
        # "Cast"
        "Pack",
        "CheckNumerics",
        "Where"
    ])


    q_type = set([
        "Dequantize",
        "QuantizeV2",
        "QuantizedConv2D",
        "QuantizedReshape",
        "RequantizationRange"
    ])

    dtype_map = {
        0 : graph_pb2.DT_UNDEFINED,
        1 : graph_pb2.DT_FLOAT32,
        2 : graph_pb2.DT_FLOAT64,
        3 : graph_pb2.DT_INT32,
        4 : graph_pb2.DT_UINT8,
        5 : graph_pb2.DT_INT16,
        6 : graph_pb2.DT_INT8,
        7 : graph_pb2.DT_STRING,
        9 : graph_pb2.DT_INT64,
        10: graph_pb2.DT_BOOL
    }


    @property
    def src_graph(self):
        return self.tf_graph

    def __init__(self, frozen_file, inputshape, in_nodes, dest_nodes):
        if LooseVersion(tensorflow.__version__) < LooseVersion('1.8.0'):
            raise ImportError(
                'Your TensorFlow version %s is outdated. '
                'MMdnn requires tensorflow>=1.8.0' % tensorflow.__version__)

        super(TensorflowParser2, self).__init__()

        self.weight_loaded = True
        # load model files into TensorFlow graph
        with open(frozen_file, 'rb') as f:
            serialized = f.read()
        tensorflow.reset_default_graph()
        original_gdef = tensorflow.GraphDef()

        original_gdef.ParseFromString(serialized)

        in_type_list = {}
        for n in original_gdef.node:
            if n.name in in_nodes:
                in_type_list[n.name] = n.attr['dtype'].type

        from tensorflow.python.tools import strip_unused_lib
        from tensorflow.python.framework import dtypes
        from tensorflow.python.platform import gfile
        original_gdef = strip_unused_lib.strip_unused(
                input_graph_def = original_gdef,
                input_node_names = in_nodes,
                output_node_names = dest_nodes,
                placeholder_type_enum = dtypes.float32.as_datatype_enum)
        # Save it to an output file
        frozen_model_file = './frozen.pb'
        with gfile.GFile(frozen_model_file, "wb") as f:
            f.write(original_gdef.SerializeToString())
        with open(frozen_model_file, 'rb') as f:
            serialized = f.read()
        tensorflow.reset_default_graph()
        model = tensorflow.GraphDef()
        model.ParseFromString(serialized)

        output_shape_map = dict()
        input_shape_map = dict()
        dtype = tensorflow.float32

        with tensorflow.Graph().as_default() as g:
            input_map = {}
            for i in range(len(inputshape)):
                if in_type_list[in_nodes[i]] == 1 or in_type_list[in_nodes[i]] == 0:
                    dtype = tensorflow.float32
                    x = tensorflow.placeholder(dtype, shape = [None] + inputshape[i])

                elif in_type_list[in_nodes[i]] == 3:
                    dtype = tensorflow.int32
                    x = tensorflow.placeholder(dtype, shape = inputshape[i])

                elif in_type_list[in_nodes[i]] == 10:
                    dtype = tensorflow.bool
                    x = tensorflow.placeholder(dtype)

                input_map[in_nodes[i] + ':0'] = x
           
            tensorflow.import_graph_def(model, name='', input_map=input_map)

        with tensorflow.Session(graph = g) as sess:

            meta_graph_def = tensorflow.train.export_meta_graph(filename='./my-model.meta')
            model = meta_graph_def.graph_def


        self.tf_graph = TensorflowGraph(model)
        self.tf_graph.build()


    @staticmethod
    def _get_scopes(layer_name):
        return layer_name.split('/')


    def check_const(self, node):
        while node:
            if node.type == "Const":
                return node
            elif node.type == "NoOp":
                return None
            else:
                node =  self.get_parent(node.name, [0])


    def _convert_reduction_operators(self, source_node, new_op = None):
        IR_node = self._convert_identity_operation(source_node, 1, new_op)

        # keep dims
        IR_node.attr['keepdims'].b = source_node.layer.attr['keep_dims'].b

        # axes
        axes = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        axes = tensor_util.MakeNdarray(axes)
        IR_node.attr['axes'].list.i.extend(axes)


    def _convert_layers_batchnorm(self, source_node):
        IR_node = self.IR_graph.node.add()
        TensorflowParser2._copy_and_reop(source_node, IR_node, 'BatchNorm')

        is_transformed = False
        test = self.get_parent(source_node.name, [0])

        if test.type == 'Mul':
            is_transformed = True

        # ssd model is transformed
        if is_transformed:
            # Ax - (Au - b)

            # A
            input_mul_A = self.get_parent(source_node.name, [0, 1])
            tensor_content = input_mul_A.get_attr('value')
            A_content = tensor_util.MakeNdarray(tensor_content)
            self.set_weight(source_node.name, 'A', A_content)

            # b
            input_sub = self.get_parent(source_node.name, [1])
            tensor_content = input_sub.get_attr('value')
            sub_content = tensor_util.MakeNdarray(tensor_content)
            # print(sub_content)
            self.set_weight(source_node.name, 'b', sub_content)

            input_node = self.get_parent(source_node.name, [0])
            IR_node.input.append(input_node.real_name)
            IR_node.attr["_output_shapes"].list.shape.pop()
            IR_node.attr["_output_shapes"].MergeFromString(input_node.layer.attr['_output_shapes'].SerializeToString())

        else:
            # epsilon
            epsilon = self.get_parent(source_node.name, [1])
            IR_node.attr['epsilon'].f = epsilon.layer.attr['value'].tensor.float_val[0]

            # moving variance (var) /read
            moving_variance = self.get_parent(source_node.name, [0])

            if moving_variance.type == 'Identity':
                moving_variance_read = self.src_graph.get_parent(moving_variance.name, [0])
                tensor_content = moving_variance_read.get_attr('value')
                moving_variance_content = tensor_util.MakeNdarray(tensor_content)
                self.set_weight(source_node.name, 'var', moving_variance_content)

            else:
                print(moving_variance.layer)
                assert False

            # gamma (scale)
            Rsqrt = self.get_son(source_node.name, [0], True)
            # print(Rsqrt.out_edges)

            if len(Rsqrt.out_edges) == 2:
                IR_node.attr['scale'].b = False
                output_node = self.get_son(Rsqrt.name, [0, 0], True)
                if output_node.type == 'Sub':
                    output_node = self.get_son(Rsqrt.name, [1, 0], True)
                    Mul = self.get_son(Rsqrt.name, [0], True)
                else:
                    Mul = self.get_son(Rsqrt.name, [1], True)
            else:
                IR_node.attr['scale'].b = True
                son = self.get_son(Rsqrt.name, [0, 0], True)
                gamma_from = self.get_parent(son.name, [1, 1], True)
                gamma = self.check_const(gamma_from)
                gamma_tensor = gamma.get_attr('value')
                scale = tensor_util.MakeNdarray(gamma_tensor)
                self.set_weight(source_node.name, 'scale', scale)
                output_node = self.get_son(source_node.name, [0, 0, 0, 0], True)
                if output_node.type == 'Sub':
                    output_node = self.get_son(source_node.name, [0, 0, 0, 0, 0], True)
                    Mul = self.get_son(Rsqrt.name, [0, 0], True)
                else:
                    Mul = self.get_son(Rsqrt.name, [0, 1], True)

            # beta  (bias)
            beta = self.get_parent(output_node.name, [1, 0, 0], True).get_attr('value')
            bias = tensor_util.MakeNdarray(beta)
            IR_node.attr['bias'].b = True
            self.set_weight(source_node.name, 'bias', bias)

            # moving mean (mean)
            moving_mean = self.get_parent(Mul.name, [0, 0]).get_attr('value')
            mean = tensor_util.MakeNdarray(moving_mean)
            self.set_weight(source_node.name, 'mean', mean)

            # input node
            assert output_node.type == 'Add'
            input_node = self.get_parent(output_node.name, [0, 0])
            IR_node.input.append(input_node.real_name)
            IR_node.attr["_output_shapes"].list.shape.pop()
            IR_node.attr["_output_shapes"].MergeFromString(input_node.layer.attr['_output_shapes'].SerializeToString())
            output_node.real_name = source_node.name


    def _convert_layers_instancenorm(self, source_node):
        IR_node = self.IR_graph.node.add()
        TensorflowParser2._copy_and_reop(source_node, IR_node, 'InstanceNorm')

        # epsilon
        epsilon = self.get_parent(source_node.name, [1])
        epsilon_value = epsilon.get_attr('value').float_val[0]
        IR_node.attr['epsilon'].f = epsilon_value

        # beta
        output_node = self.get_son(source_node.name, [0, 0, 0, 0], True)
        beta = self.get_parent(output_node.name, [1, 0, 0, 0, 0, 1], True)
        beta_tensor = beta.get_attr('value')
        beta = tensor_util.MakeNdarray(beta_tensor)
        self.set_weight(source_node.name, 'bias', beta)


        # gamma (scale)
        IR_node.attr['scale'].b = True
        son = self.get_son(source_node.name, [0, 0, 0], True)
        gamma = self.get_parent(son.name, [1, 1, 0, 0, 0, 1], True)
        gamma_tensor = gamma.get_attr('value')
        scale = tensor_util.MakeNdarray(gamma_tensor)
        self.set_weight(source_node.name, 'scale', scale)
        # output_node = self.get_son(source_node.name, [0, 0, 0, 0], True)

        assert output_node.type == 'Add'
        input_node = self.get_parent(output_node.name, [0, 0])
        IR_node.input.append(input_node.real_name)

        output_node.real_name = source_node.name

        # assert False


    @classmethod
    def _skip_node(cls, source_node):
        if source_node.covered:
            return True

        for prefix in cls.skip_prefix:
            if source_node.name.startswith(prefix):
                return True

        scopes = TensorflowParser2._get_scopes(source_node.name)

        for s in scopes:
            if s in cls.skip_scope:
                return True

        return False


    def _add_constant_node(self, source_node):
        parent_ids=range(len(source_node.in_edges))
        for idx in parent_ids:
            s = source_node.in_edges[idx]
            parent_node = self.tf_graph.get_node(s)
            if parent_node.type == 'Const':
                self._rename_Const(parent_node)
    
    def _rename_Const(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=0, new_op='Constant') # Constant
        value = source_node.get_attr('value')
        if value.float_val:
            value = value.float_val[0]
        elif value.int_val:
            value = value.int_val[0]
        else:
            value = tensor_util.MakeNdarray(value).tolist()
        kwargs = {'value': value}
        assign_IRnode_values(IR_node, kwargs)


    def gen_IR(self):

        for layer in self.src_graph.topological_sort:
            current_node = self.src_graph.get_node(layer)

            if self._skip_node(current_node):
                continue

            node_type = current_node.type
            
            if hasattr(self, "rename_" + node_type):
 
                func = getattr(self, "rename_" + node_type)
                func(current_node)
            else:
                
                self.rename_UNKNOWN(current_node)


    @staticmethod
    def tensor_shape_to_list(shapes):
        if isinstance(shapes, attr_value_pb2.AttrValue):
            return [dim.size for dim in shapes.shape.dim]

        else:
            ret = []
            for shape in shapes:
                this_one = [dim.size for dim in shape.dim]
                ret.append(this_one)
            return ret

    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op = None):
        if new_op == None: new_op = source_node.type
        IR_node.name = source_node.name
        IR_node.op = new_op

        kwargs = {}
        if 'data_format' in source_node.layer.attr:
            kwargs['data_format'] = source_node.get_attr('data_format')

        if 'T' in source_node.layer.attr:
            if source_node.type not in TensorflowParser2.q_type:
                if source_node.type == 'Enter':
                    IR_node.attr["dtype"].type = TensorflowParser2.dtype_map[6]
                else:
                    assert source_node.layer.attr['T'].type in TensorflowParser2.dtype_map, 'type [{}] is unknown.'.format(source_node.layer.attr['dtype'].type)
                    IR_node.attr["dtype"].type = TensorflowParser2.dtype_map[source_node.layer.attr['T'].type]
            else:
                # Quantized model type
                IR_node.attr["dtype"].type = TensorflowParser2.dtype_map[6]

        if '_output_shapes' in source_node.layer.attr:
            IR_node.attr["_output_shapes"].MergeFromString(source_node.layer.attr['_output_shapes'].SerializeToString())

        if 'paddings' in source_node.layer.attr:
            IR_node.attr["paddings"].MergeFromString(source_node.layer.attr['paddings'].SerializeToString())

        assign_IRnode_values(IR_node, kwargs)

    def _convert_inedge(self, source_node, IR_node, start_idx = 0, end_idx = None):
        if end_idx == None: end_idx = len(source_node.in_edges)
        for idx in range(start_idx, end_idx):
            IR_node.input.append(self.src_graph.get_node(source_node.in_edges[idx]).real_name)

    @staticmethod
    def _copy_shape(source_node, IR_node):
        assert 'shape' in source_node.layer.attr
        if source_node.layer.attr['shape'].list.shape:
            IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['shape'].list.shape[0].SerializeToString())
        else:
            IR_node.attr['shape'].shape.MergeFromString(source_node.layer.attr['shape'].shape.SerializeToString())

    def rename_UNKNOWN(self, source_node):
        if source_node.type in self.skip_type:
            return
        print("Tensorflow has not supported operator [%s] with name [%s]."
              % (source_node.type, source_node.name))
        assert False

    def rename_NoOp(self, source_node):
        return

    def _convert_padding(self, source_node, IR_node, kernel_size):
        # TODO: Fused conv and pool with padding is different from defused operators
        input_node = self.get_parent(source_node.name, [0])
        input_shape = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        if source_node.get_attr('padding') == 'VALID':
            dims = len(input_shape)
            assign_IRnode_values(IR_node, {'auto_pad' : "VALID", 'pads' : [0, 0] * dims})

        elif source_node.get_attr('padding') == 'SAME':
            padding = compute_tf_same_padding(
                input_shape,
                kernel_size,
                source_node.get_attr('strides'))
            assign_IRnode_values(IR_node, {'auto_pad' : "SAME_LOWER", 'pads' : padding})

        else:
            assert False


    def _get_bias(self, source_node, IR_node):
        if not source_node.out_edges:
            return

        add_node = self.tf_graph.get_node(source_node.out_edges[0])
        if add_node.type != "Add" and add_node.type != "BiasAdd":
            return

        variable = self.check_const(self.tf_graph.get_node(add_node.in_edges[1])) #add_bias node
        if variable.type != 'Const':
            return


        bias_value = variable.get_attr('value')
        bias = tensor_util.MakeNdarray(bias_value)

        # assert variable.get_attr('_output_shapes')[0].dim[0].size == IR_node.attr['kernel_shape'].list.i[-1]


        add_node.real_name = IR_node.name
        add_node.covered = True
        IR_node.attr['use_bias'].b = True
        current_layer = self.weights[source_node.name]
        current_layer['bias'] = bias


    def _convert_pooling(self, source_node, pool_type):

        IR_node = self._convert_identity_operation(source_node, new_op='Pool')
        kwargs = {}

        # strides
        kwargs['strides'] = source_node.get_attr('strides')

        # window_shape
        kwargs['kernel_shape'] = source_node.get_attr('ksize')

        # pool type
        kwargs['pooling_type'] = pool_type

        # padding
        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][1:-1])


        assign_IRnode_values(IR_node, kwargs)


    def _convert_identity_operation(self, source_node, start_idx = 0, end_idx = None, new_op = None):
        IR_node = self.IR_graph.node.add()
        TensorflowParser2._copy_and_reop(source_node, IR_node, new_op)
        self._convert_inedge(source_node, IR_node, start_idx, end_idx)
        return IR_node


    def rename_Relu6(self, source_node):
        self._convert_identity_operation(source_node, new_op = 'Relu6')


    def rename_Merge(self, source_node):
        # In facenet or other newtwork using slim.batch_norm,
        # There are two BN(train, test) skip switch and merge.
        source_node.real_name =  self.src_graph.get_node(source_node.in_edges[0]).real_name


    def rename_DepthwiseConv2dNative(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx = 1, new_op = 'DepthwiseConv')
        kwargs = {}
        kwargs['strides'] = source_node.get_attr('strides')
        input_node = self.src_graph.get_parent(source_node.name, [1])
        kwargs['kernel_shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][:-2])


        weight_node = self.src_graph.get_parent(source_node.name, [1])
        weight = self.check_const(weight_node).get_attr('value')
        weight_content = tensor_util.MakeNdarray(weight)
        self.set_weight(source_node.name, 'weights', weight_content)
        assign_IRnode_values(IR_node, kwargs)


    def rename_BatchNormWithGlobalNormalization(self, source_node):

        IR_node = self._convert_identity_operation(source_node, start_idx=0, end_idx=1, new_op='BatchNorm')
        # epsilon
        IR_node.attr['epsilon'].f = source_node.get_attr('variance_epsilon')

        # moving variance (var) /read
        moving_variance = self.get_parent(source_node.name, [2])
        tensor_variance = moving_variance.get_attr('value')
        moving_variance_content = tensor_util.MakeNdarray(tensor_variance)
        self.set_weight(source_node.name, 'var', moving_variance_content)

        # gamma (scale)
        gamma = self.get_parent(source_node.name, [4])
        gamma_value = gamma.get_attr('value')
        gamma = tensor_util.MakeNdarray(gamma_value)
        self.set_weight(source_node.name, 'scale', gamma)
        IR_node.attr['scale'].b = True

        # beta  (bias)
        beta = self.get_parent(source_node.name, [3])
        beta_value = beta.get_attr('value')
        beta = tensor_util.MakeNdarray(beta_value)
        self.set_weight(source_node.name, 'bias', beta)
        IR_node.attr['use_bias'].b = True

        # moving mean (mean)
        mean = self.get_parent(source_node.name, [1])
        mean_value = mean.get_attr('value')
        mean = tensor_util.MakeNdarray(mean_value)
        self.set_weight(source_node.name, 'mean', mean)

    def rename_Placeholder(self, source_node):
        if source_node.layer.attr["shape"].shape.unknown_rank == True:
            return
        IR_node = self._convert_identity_operation(source_node, new_op='DataInput')
        TensorflowParser2._copy_shape(source_node, IR_node)
        IR_node.attr['shape'].shape.dim[0].size = -1
        IR_node.attr['_output_shapes'].list.shape[0].dim[0].size = -1


    def rename_Mean(self, source_node):
        # ReduceMean
        IR_node = self._convert_identity_operation(source_node, start_idx = 0, end_idx = 1, new_op='ReduceMean')
        # keep dims
        IR_node.attr['keepdims'].b = source_node.layer.attr['keep_dims'].b

        # axes
        axes = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        axes = tensor_util.MakeNdarray(axes)
        IR_node.attr['axes'].list.i.extend(axes)


    def rename_Reshape(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx = 1)
        kwargs = {'shape' : self.tensor_shape_to_list(source_node.get_attr('_output_shapes'))[0]}
        assign_IRnode_values(IR_node, kwargs)


    def rename_MirrorPad(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'MirrorPad')
        input_node = self.src_graph.get_parent(source_node.name, [1])

        tensor_content = tensor_util.MakeNdarray(input_node.get_attr('value')).reshape(-1)
        kwargs = {}
        kwargs['mode'] = source_node.get_attr('mode')
        kwargs['pads'] = tensor_content.tolist()

        assign_IRnode_values(IR_node, kwargs)


    def rename_Min(self, source_node):
        IR_node = self._convert_identity_operation(source_node, start_idx=0, end_idx=1, new_op = 'Min')
        kwargs = {}
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape_0'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        input_node = self.src_graph.get_parent(source_node.name, [1])
        kwargs['shape_1'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]
        assign_IRnode_values(IR_node, kwargs)


    def rename_Max(self, source_node):
        IR_node = self._convert_identity_operation(source_node, start_idx=0, end_idx=1, new_op = 'Max')
        kwargs = {}
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape_0'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        input_node = self.src_graph.get_parent(source_node.name, [1])
        kwargs['shape_1'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]
        assign_IRnode_values(IR_node, kwargs)


    def rename_Mul(self, source_node):
        scopes = self._get_scopes(source_node.name)

        if len(scopes) >= 2:
            if scopes[-2] == "batchnorm" or scopes[-2].startswith("Assign"):
                return
        self._add_constant_node(source_node)
        self._convert_identity_operation(source_node)


    def rename_Add(self, source_node):
        scopes = self._get_scopes(source_node.name)
        if len(scopes) > 2:
            if scopes[-2] == 'batchnorm':
                if scopes[-3] == 'BatchNorm' or scopes[-3] == 'batch_normalization':
                    self._convert_layers_batchnorm(source_node)
                elif scopes[-3] == 'InstanceNorm':
                    self._convert_layers_instancenorm(source_node)
            else:
                IR_node = self._convert_identity_operation(source_node, new_op = "Add")
        else:
            IR_node = self._convert_identity_operation(source_node, new_op = "Add")


    def rename_Fill(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Fill")


    def rename_Sub(self, source_node):
        scopes = self._get_scopes(source_node.name)
        if len(scopes) > 2:
            if scopes[-2].startswith('Assign') or scopes[-1].startswith('Assign'):
                return
        IR_node = self._convert_identity_operation(source_node, end_idx=2, new_op = "Sub")


    def rename_Sum(self, source_node):
        IR_node = self._convert_identity_operation(source_node, start_idx=0, end_idx=1, new_op = 'Sum')
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs = {}
        kwargs['cal_shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        input_node_indices = self.src_graph.get_parent(source_node.name, [1])
        indice_value = input_node_indices.get_attr('value')
        if indice_value.tensor_content:
            shapes = tensor_util.MakeNdarray(indice_value)
            c = shapes.tolist()
            kwargs['sum_indices'] = c
        else:
            kwargs['sum_indices'] = input_node_indices.get_attr('value').int_val[0]
        assign_IRnode_values(IR_node, kwargs)

    def rename_Rsqrt(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = "Rsqrt")

        kwargs = {}
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)

    def rename_Square(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Square')
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs = {}
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)


    def rename_Sigmoid(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        kwargs = {}
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)

    def rename_Reciprocal(self, source_node):
        IR_node = self._convert_identity_operation(source_node, start_idx=0, end_idx=1, new_op = 'Reciprocal')


    def rename_Minimum(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Minimum')

    def rename_Maximum(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Maximum')

    def rename_RealDiv(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'RealDiv')

    def rename_Enter(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Enter')

    def rename_Switch(self, source_node):
        # Skip the node as merge
        source_node.real_name =  self.src_graph.get_node(source_node.in_edges[0]).real_name


    def rename_Identity(self, source_node):
        source_node.real_name =  self.src_graph.get_node(source_node.in_edges[0]).real_name


    def rename_Exp(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Exp')

    def rename_ResizeBilinear(self, source_node):
        IR_node = self._convert_identity_operation(source_node,end_idx=1, new_op = 'ResizeBilinear')


    def rename_Cast(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Cast')
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs = {}
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)


    def rename_Prod(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Prod')

        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs = {}
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        input_node_const = self.src_graph.get_parent(source_node.name, [1])

        kwargs['const'] = input_node_const.get_attr('value').int_val[0]
        assign_IRnode_values(IR_node, kwargs)


    def rename_Shape(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Shape')
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs = {}
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)


    def rename_Squeeze(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Squeeze')


    def rename_Gather(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Embedding')

        W = self.src_graph.get_parent(source_node.name, [0])
        W = self.src_graph.get_parent(W.name, [0])

        self.set_weight(source_node.name, "weights", self.ckpt_data[W.name])

        kwargs = {
            'input_dim' : self.ckpt_data[W.name].shape[0],
            'output_dim' : self.ckpt_data[W.name].shape[1],
            'mask_zero' : False
        }
        kwargs['axis'] = 0  # add default
        assign_IRnode_values(IR_node, kwargs)

        return IR_node
    
    def rename_GatherV2(self, source_node):
        
        IR_node = self.rename_Gather(source_node)

        kwargs = {}
        kwargs['axis'] = source_node.layer.attr['axis'].i
        assign_IRnode_values(IR_node, kwargs)


    def rename_StridedSlice(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'Slice')
        kwargs = {}
        kwargs = {
            'begin_mask' : source_node.get_attr('begin_mask'),
            'end_mask'   : source_node.get_attr('end_mask'),
        }

        starts = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        starts = tensor_util.MakeNdarray(starts).tolist()
        kwargs['starts'] = starts

        ends = self.get_parent(source_node.name, [2]).layer.attr['value'].tensor
        ends = tensor_util.MakeNdarray(ends).tolist()
        kwargs['ends'] = ends

        if self.get_parent(source_node.name, [3]) != None:
            strides = self.get_parent(source_node.name, [3]).layer.attr['value'].tensor
            strides = tensor_util.MakeNdarray(strides).tolist()
            kwargs['strides'] = strides

        assign_IRnode_values(IR_node, kwargs)


    def rename_ResizeNearestNeighbor(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        kwargs = {}
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        input_node_size = self.src_graph.get_parent(source_node.name, [1])
        kwargs['size'] = self.tensor_shape_to_list(input_node_size.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)


    def rename_Conv2D(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'Conv')
        kwargs = {}
        kwargs['strides'] = source_node.get_attr('strides')
        kwargs['padding'] = source_node.get_attr('padding')

        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        # weights
        input_node_weight = self.src_graph.get_parent(source_node.name, [1])
        tensor_content = self.check_const(input_node_weight).get_attr('value')
        W = tensor_util.MakeNdarray(tensor_content)

        kwargs['kernel_shape'] = self.tensor_shape_to_list(input_node_weight.get_attr('_output_shapes'))[0]

        self.set_weight(source_node.name, 'weights', W)

        self._convert_padding(source_node, IR_node, kwargs['kernel_shape'][:-2])

        assign_IRnode_values(IR_node, kwargs)
        self._get_bias(source_node, IR_node)


    def rename_Relu(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        kwargs = {'shape' : self.tensor_shape_to_list(source_node.get_attr('_output_shapes'))[0]}
        assign_IRnode_values(IR_node, kwargs)


    def rename_MaxPool(self, source_node):
        self._convert_pooling(source_node, b'MAX')


    def rename_AvgPool(self, source_node):
        self._convert_pooling(source_node, b'AVG')


    def rename_LRN(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        # alpha
        IR_node.attr["alpha"].f = float(source_node.get_attr("alpha", "0.0001"))
        # beta
        IR_node.attr["beta"].f = float(source_node.get_attr("beta", "0.75"))
        IR_node.attr["size"].i = source_node.get_attr("depth_radius")
        IR_node.attr["bias"].f = float(source_node.get_attr("bias"))


    def rename_Concat(self, source_node):
        n = len(source_node.in_edges)
        IR_node = self._convert_identity_operation(source_node, start_idx=1, end_idx=n, new_op='Concat')
        axis = self.tf_graph.get_parent(source_node.name, [0])
        IR_node.attr["axis"].i = axis.get_attr('value').int_val[0]


    def rename_ConcatV2(self, source_node):
        n = len(source_node.in_edges)
        IR_node = self._convert_identity_operation(source_node, start_idx=0, end_idx=n-1, new_op='Concat')
        axis = self.tf_graph.get_parent(source_node.name, [n-1])
        IR_node.attr["axis"].i = axis.get_attr('value').int_val[0]


    def rename_MatMul(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1)
        input_weight_node = self.src_graph.get_parent(source_node.name, [1])
        weightnode = self.check_const(input_weight_node)
        weight_value = weightnode.get_attr('value')

        weight = tensor_util.MakeNdarray(weight_value)
        self.set_weight(source_node.name, 'weights', weight)

        units = source_node.layer.attr['_output_shapes'].list.shape[-1].dim[-1].size
        IR_node.attr['units'].i = units

        if source_node.out_edges and self.tf_graph.get_node(source_node.out_edges[0]).type == 'BiasAdd':
            add_node = self.tf_graph.get_node(source_node.out_edges[0])
            add_node.covered = True
            add_node.real_name = source_node.real_name

            TensorflowParser2._copy_and_reop(source_node, IR_node, 'FullyConnected')
            variable = self.tf_graph.get_node(add_node.in_edges[1]) #add_bias node
            biasnode = self.check_const(variable)
            bias_value = biasnode.get_attr('value')
            bias = tensor_util.MakeNdarray(bias_value)
            self.set_weight(source_node.name, 'bias', bias)
            IR_node.attr['use_bias'].b = True


    def rename_Softmax(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        kwargs = {'shape' : self.tensor_shape_to_list(source_node.get_attr('_output_shapes'))[0]}
        IR_node.attr["dim"].i = 1
        assign_IRnode_values(IR_node, kwargs)


    def rename_BiasAdd(self, source_node):
        # Skip BiasAdd
        source_node.real_name =  self.src_graph.get_node(source_node.in_edges[0]).real_name


    def rename_QuantizeV2(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'QuantizeV2')
        TensorflowParser2._copy_shape(source_node, IR_node)


    def rename_QuantizedRelu(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = "QuantizedRelu")
        kwargs = {'shape' : self.tensor_shape_to_list(source_node.get_attr('_output_shapes'))[0]}
        assign_IRnode_values(IR_node, kwargs)


    def rename_QuantizedReshape(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx = 1)
        kwargs = {'shape' : self.tensor_shape_to_list(source_node.get_attr('_output_shapes'))[0]}
        assign_IRnode_values(IR_node, kwargs)


    def rename_QuantizedConv2D(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'QConv')
        kwargs = {}
        kwargs['strides'] = source_node.get_attr('strides')
        kwargs['padding'] = source_node.get_attr('padding')

        # weights
        input_node = self.src_graph.get_parent(source_node.name, [1])
        tensor_content = input_node.get_attr('value')
        W = tensor_util.MakeNdarray(tensor_content)
        W = W.astype(np.uint8)

        kwargs['kernel_shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        input_node_minw = self.src_graph.get_parent(source_node.name, [4])
        min_W = input_node_minw.get_attr('value').float_val[0]
        input_node_maxw = self.src_graph.get_parent(source_node.name, [5])
        max_W = input_node_maxw.get_attr('value').float_val[0]

        if source_node.get_attr('Tfilter') == tensorflow.quint8:
            W = ((max_W - min_W)/255.0) * W + min_W
        else:
            assert False, ('Only uint8 weights handled currently by the converter')

        self.set_weight(source_node.name, 'kernel_weights', W)
        assign_IRnode_values(IR_node, kwargs)


    def rename_Requantize(self, source_node):
        input_node = self.get_parent(source_node.name, [0])
        son_node = self.get_son(source_node.name, [0])

        son_node.real_name = source_node.name

    def rename_RequantizationRange(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'RequantizationRange')
        TensorflowParser2._copy_shape(source_node, IR_node)


    def rename_ZerosLike(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'ZerosLike')


    def rename_Rank(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Rank')


    def rename_Transpose(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'Transpose')
        
        input_node_perm = self.get_parent(source_node.name, [1])
        # input_node_perm = self.check_const(self.get_parent(source_node.name, [1], True))
        tensor_content = input_node_perm.get_attr('value')
        perm = tensor_util.MakeNdarray(tensor_content).tolist()
        assign_IRnode_values(IR_node, {'perm' : perm})

    def rename_GreaterEqual(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'GreaterEqual')


    def rename_Greater(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'Greater')


    def rename_Equal(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Equal')


    def rename_All(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'All')


    def rename_LogicalAnd(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Mul')


    def rename_Pad(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'Pad')
        kwargs = {}
        kwargs['mode'] = 'constant'

        # paddings
        padding = self.get_parent(source_node.name, [1]).layer.attr['value'].tensor
        shapes = tensor_util.MakeNdarray(padding)
        kwargs['pads'] = convert_tf_pad_to_onnx(shapes)

        assign_IRnode_values(IR_node, kwargs)


    def rename_FusedBatchNorm(self, source_node):
        scalenode = self.check_const(self.get_parent(source_node.name, [1], True))
        if ':' in source_node.in_edges[1]: # ?
            scalenode = None

        if scalenode:
            scale_value = scalenode.get_attr('value')
            IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'BatchNorm')
            # for attr.shape >= 2
            for i in range(len(IR_node.attr["_output_shapes"].list.shape)-1):
                IR_node.attr["_output_shapes"].list.shape.pop()

        else:
            # For models built by slim.batch_norm, remove duplicate BN (eg.facenet)
            return

        scale = tensor_util.MakeNdarray(scale_value)
        self.set_weight(source_node.name, 'scale', scale)
        IR_node.attr['scale'].b = True


        IR_node.attr['epsilon'].f = source_node.get_attr('epsilon', 0)
        biasnode = self.check_const(self.get_parent(source_node.name, [2], True))
        if biasnode:
            bias_value = biasnode.get_attr('value')
        else:
            innode = self.get_parent(source_node.name, [2], True)
            name = innode.name.split(':')[0]
            bias_value = self.check_const(self.src_graph.layer_map[name]).get_attr('value')
        bias = tensor_util.MakeNdarray(bias_value)
        self.set_weight(source_node.name, 'bias', bias)
        IR_node.attr['bias'].b = True

        meannode = self.check_const(self.get_parent(source_node.name, [3], True))
        mean_value = meannode.get_attr('value')
        mean = tensor_util.MakeNdarray(mean_value)
        self.set_weight(source_node.name, 'mean', mean)

        variancenode = self.check_const(self.get_parent(source_node.name, [4], True))
        variance_value = variancenode.get_attr('value')
        variance = tensor_util.MakeNdarray(variance_value)
        self.set_weight(source_node.name, 'var', variance)


    def rename_SpaceToBatchND(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'SpaceToBatchND')


    def rename_BatchToSpaceND(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'BatchToSpaceND')


    def rename_ArgMax(self, source_node):
        IR_node = self._convert_identity_operation(source_node, end_idx=1, new_op = 'ArgMax')


    def rename_Slice(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Slice')


    def rename_Split(self, source_node):
        if source_node.get_attr('num_split') == 1:
            source_node.real_name = self.get_parent(source_node.name, [1]).real_name

        else:
            IR_node = self._convert_identity_operation(source_node, start_idx=1, new_op = 'Split')
            kwargs = {
                'axis' : self.get_parent(source_node.name, [0]).layer.attr['value'].tensor.int_val[0],
                'split' : source_node.get_attr('num_split')
            }
            assign_IRnode_values(IR_node, kwargs)


    def rename_Tile(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Tile')


    def rename_Sqrt(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Sqrt')


    def rename_Tanh(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        kwargs = {}
        input_node = self.src_graph.get_parent(source_node.name, [0])
        kwargs['shape'] = self.tensor_shape_to_list(input_node.get_attr('_output_shapes'))[0]

        assign_IRnode_values(IR_node, kwargs)
    
    def rename_Log(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Log')


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import numpy as np
import sys
import os
from mmdnn.conversion.examples.imagenet_test import TestKit

import colorsys
from keras import backend as K
from PIL import Image, ImageFont, ImageDraw
from mmdnn.conversion.examples.keras.utils import yolo_eval


class TestKeras(TestKit):

    def __init__(self):

        # self.anchors = np.array([[10,13], [16,30],[33,23],[30,61],[62,45], [59,119],[116,90],[156,198],[373,326]])
        self.class_names = ['person', 'bicycle', 'car', 'motorbike', 'aeroplane', 'bus', 'train', 'truck', 'boat',
        'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow',
        'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard',
        'sports ball', 'kite', 'baseball bat', 'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup',
        'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza', 'donut',
        'cake', 'chair', 'sofa', 'pottedplant', 'bed', 'diningtable', 'toilet', 'tvmonitor', 'laptop', 'mouse', 'remote', 'keyboard',
        'cell phone', 'microwave', 'oven', 'toaster', 'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush']
        super(TestKeras, self).__init__()
        self.model = self.MainModel.KitModel(self.args.w)


    def preprocess(self, image_path):
        x = super(TestKeras, self).preprocess(image_path)
        self.data = np.expand_dims(x, 0)

    def print_result(self):
        predict = self.model.predict(self.data)
        super(TestKeras, self).print_result(predict)

    def generate(self):
        self.input_image_shape = K.placeholder(shape=(2, ))
        output = self.model.output
        output.sort(key=lambda x: int(x.shape[1]))
        # print(output)

        boxes, scores, classes = yolo_eval(output, self.anchors,
                    len(self.class_names), self.input_image_shape,
                    score_threshold=self.score_threshold, iou_threshold=self.iou_threshold)
        return boxes, scores, classes

    def yolo_result(self, path):
        image = Image.fromarray(np.uint8(np.squeeze(self.data)))

        self.sess = K.get_session()
        self.boxes, self.scores, self.classes = self.generate()
        out_boxes, out_scores, out_classes = self.sess.run(
            [self.boxes, self.scores, self.classes],
            feed_dict={
                self.model.input: self.data/255.,
                self.input_image_shape: [608, 608],
                K.learning_phase(): 0
            })
        # print(out_boxes, out_scores, out_classes)
        print('Found {} boxes for {}'.format(len(out_boxes), 'img'))

        thickness = (image.size[0] + image.size[1]) // 300

        for i, c in reversed(list(enumerate(out_classes))):
            predicted_class = self.class_names[c]
            box = out_boxes[i]
            score = out_scores[i]

            label = '{} {:.2f}'.format(predicted_class, score)
            draw = ImageDraw.Draw(image)
            label_size = draw.textsize(label)

            top, left, bottom, right = box
            top = max(0, np.floor(top + 0.5).astype('int32'))
            left = max(0, np.floor(left + 0.5).astype('int32'))
            bottom = min(image.size[1], np.floor(bottom + 0.5).astype('int32'))
            right = min(image.size[0], np.floor(right + 0.5).astype('int32'))
            print(label, (left, top), (right, bottom))

            if top - label_size[1] >= 0:
                text_origin = np.array([left, top - label_size[1]])
            else:
                text_origin = np.array([left, top + 1])

            # get random colors
            self.colors = []
            C = list(np.random.random_integers(255, size=(len(self.class_names),3)))
            for i in C:
                self.colors.append(tuple(i))

            # My kingdom for a good redistributable image drawing library.
            for i in range(thickness):
                draw.rectangle(
                    [left + i, top + i, right - i, bottom - i],
                    outline=self.colors[c])
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)],
                fill=self.colors[c])
            draw.text(tuple(text_origin), label, fill=(0, 0, 0))
            del draw
        image.save("{}.jpg".format(path), "JPEG")


    def print_intermediate_result(self, layer_name, if_transpose = False):
        from keras.models import Model
        intermediate_layer_model = Model(inputs = self.model.input,
                                         outputs = self.model.get_layer(layer_name).output)
        intermediate_output = intermediate_layer_model.predict(self.data)
        super(TestKeras, self).print_intermediate_result(intermediate_output, if_transpose)


    def inference(self, image_path):
        self.preprocess(image_path)

        print(self.data.shape)
        # self.print_intermediate_result('conv1_7x7_s2_1', True)

        self.print_result()

        self.test_truth()

    def dump(self, path = None):
        if path is None: path = self.args.dump

        self.model.save(path)
        print ('Keras model file is saved as [{}], generated by [{}.py] and [{}].'.format(
            path, self.args.n, self.args.w))

    def detect(self, image_path, path = None):
        self.yolo_parameter = self.MainModel.yolo_parameter()
        # yolov3 80 classes
        assert self.yolo_parameter[1] == 80
        self.anchors = []
        for i in range(len(self.yolo_parameter[0])):
            if i%2:
                tmp = [self.yolo_parameter[0][i-1], self.yolo_parameter[0][i]]
                self.anchors.append(tmp)
        self.anchors = np.array(self.anchors)
        self.score_threshold = self.yolo_parameter[2]
        self.iou_threshold = self.yolo_parameter[3]

        self.preprocess(image_path)

        self.yolo_result(path)

        print ('Keras yolo model result file is saved as [{}.jpg], generated by [{}.py] and [{}].'.format(
            path, self.args.n, self.args.w))


if __name__=='__main__':
    tester = TestKeras()
    if tester.args.dump:
        tester.dump()
    elif tester.args.detect:
        tester.detect(tester.args.image, tester.args.detect)
    else:
        tester.inference(tester.args.image)



from mmdnn.conversion.rewriter.rewriter import UnitRewriterBase
from mmdnn.conversion.tensorflow.rewriter.gru_rewriter import GRURewriter
from mmdnn.conversion.tensorflow.rewriter.lstm_rewriter import LSTMRewriter

def process_graph(graph, weights):
    rewriter_list = [GRURewriter, LSTMRewriter]

    for rewriter in rewriter_list:
        rewriter(graph, weights).run()


import numpy as np
import os
import sys
import tensorflow as tf
import tensorflow.contrib.slim as slim

def create_symbol(X, num_classes=0, is_training=False, CUDNN=False, 
                  maxf=30000, edim=125, nhid=100, batchs=64):
    word_vectors = tf.contrib.layers.embed_sequence(X, vocab_size=maxf, embed_dim=edim)
    

    word_list = tf.unstack(word_vectors, axis=1)
    
    if not CUDNN:
        cell1 = tf.contrib.rnn.LSTMCell(nhid)
        cell2 = tf.contrib.rnn.GRUCell(nhid)
        stacked_cell = tf.nn.rnn_cell.MultiRNNCell([cell1, cell2])
        outputs, states = tf.nn.static_rnn(stacked_cell, word_list, dtype=tf.float32)
        logits = tf.layers.dense(outputs[-1], 2, activation=None, name='output')
    else:
        # Using cuDNN since vanilla RNN
        from tensorflow.contrib.cudnn_rnn.python.ops import cudnn_rnn_ops
        cudnn_cell = cudnn_rnn_ops.CudnnGRU(num_layers=1, 
                                            num_units=nhid, 
                                            input_size=edim, 
                                            input_mode='linear_input')
        params_size_t = cudnn_cell.params_size()
        params = tf.Variable(tf.random_uniform([params_size_t], -0.1, 0.1), validate_shape=False)   
        input_h = tf.Variable(tf.zeros([1, batchs, nhid]))
        outputs, states = cudnn_cell(input_data=word_list,
                                     input_h=input_h,
                                     params=params)
        logits = tf.layers.dense(outputs[-1], 2, activation=None, name='output')
    
    return logits, logits

def dummy_arg_scope():
    with slim.arg_scope([]) as sc:
        return sc


import sys as _sys
import google.protobuf.text_format as text_format
from six import text_type as _text_type


def _convert(args):
    if args.inputShape != None:
        inputshape = []
        for x in args.inputShape:
            shape = x.split(',')
            inputshape.append([int(x) for x in shape])
    else:
        inputshape = [None]
    if args.srcFramework == 'caffe':
        from mmdnn.conversion.caffe.transformer import CaffeTransformer
        transformer = CaffeTransformer(args.network, args.weights, "tensorflow", inputshape[0], phase = args.caffePhase)
        graph = transformer.transform_graph()
        data = transformer.transform_data()

        from mmdnn.conversion.caffe.writer import JsonFormatter, ModelSaver, PyWriter
        JsonFormatter(graph).dump(args.dstPath + ".json")
        print ("IR network structure is saved as [{}.json].".format(args.dstPath))

        prototxt = graph.as_graph_def().SerializeToString()
        with open(args.dstPath + ".pb", 'wb') as of:
            of.write(prototxt)
        print ("IR network structure is saved as [{}.pb].".format(args.dstPath))

        import numpy as np
        with open(args.dstPath + ".npy", 'wb') as of:
            np.save(of, data)
        print ("IR weights are saved as [{}.npy].".format(args.dstPath))

        return 0

    elif args.srcFramework == 'caffe2':
        raise NotImplementedError("Caffe2 is not supported yet.")

    elif args.srcFramework == 'keras':
        if args.network != None:
            model = (args.network, args.weights)
        else:
            model = args.weights

        from mmdnn.conversion.keras.keras2_parser import Keras2Parser
        parser = Keras2Parser(model)

    elif args.srcFramework == 'tensorflow' or args.srcFramework == 'tf':
        assert args.network or args.weights
        if not args.network:
            if args.dstNodeName is None:
                raise ValueError("Need to provide the output node of Tensorflow model.")
            if args.inNodeName is None:
                raise ValueError("Need to provide the input node of Tensorflow model.")
            if inputshape is None:
                raise ValueError("Need to provide the input node shape of Tensorflow model.")
            assert len(args.inNodeName) == len(inputshape)
            from mmdnn.conversion.tensorflow.tensorflow_frozenparser import TensorflowParser2
            parser = TensorflowParser2(args.weights, inputshape, args.inNodeName, args.dstNodeName)

        else:
            from mmdnn.conversion.tensorflow.tensorflow_parser import TensorflowParser
            if args.inNodeName and inputshape[0]:
                parser = TensorflowParser(args.network, args.weights, args.dstNodeName, inputshape[0], args.inNodeName)
            else:
                parser = TensorflowParser(args.network, args.weights, args.dstNodeName)

    elif args.srcFramework == 'mxnet':
        assert inputshape != None
        if args.weights == None:
            model = (args.network, inputshape[0])
        else:
            import re
            if re.search('.', args.weights):
                args.weights = args.weights[:-7]
            prefix, epoch = args.weights.rsplit('-', 1)
            model = (args.network, prefix, epoch, inputshape[0])

        from mmdnn.conversion.mxnet.mxnet_parser import MXNetParser
        parser = MXNetParser(model)

    elif args.srcFramework == 'cntk':
        from mmdnn.conversion.cntk.cntk_parser import CntkParser
        model = args.network or args.weights
        parser = CntkParser(model)

    elif args.srcFramework == 'pytorch':
        assert inputshape != None
        from mmdnn.conversion.pytorch.pytorch_parser import PytorchParser
        model = args.network or args.weights
        assert model != None
        parser = PytorchParser(model, inputshape[0])

    elif args.srcFramework == 'torch' or args.srcFramework == 'torch7':
        from mmdnn.conversion.torch.torch_parser import TorchParser
        model = args.network or args.weights
        assert model != None
        parser = TorchParser(model, inputshape[0])

    elif args.srcFramework == 'onnx':
        from mmdnn.conversion.onnx.onnx_parser import ONNXParser
        parser = ONNXParser(args.network)

    elif args.srcFramework == 'darknet':
        from mmdnn.conversion.darknet.darknet_parser import DarknetParser
        parser = DarknetParser(args.network, args.weights, args.darknetStart)

    elif args.srcFramework == 'coreml':
        from mmdnn.conversion.coreml.coreml_parser import CoremlParser
        parser = CoremlParser(args.network)

    else:
        raise ValueError("Unknown framework [{}].".format(args.srcFramework))

    parser.run(args.dstPath)

    return 0


def _get_parser():
    import argparse

    parser = argparse.ArgumentParser(description = 'Convert other model file formats to IR format.')

    parser.add_argument(
        '--srcFramework', '-f',
        type=_text_type,
        choices=["caffe", "caffe2", "cntk", "mxnet", "keras", "tensorflow", 'tf', 'torch', 'torch7', 'onnx', 'darknet', 'coreml', 'pytorch'],
        help="Source toolkit name of the model to be converted.")

    parser.add_argument(
        '--weights', '-w', '-iw',
        type=_text_type,
        default=None,
        help='Path to the model weights file of the external tool (e.g caffe weights proto binary, keras h5 binary')

    parser.add_argument(
        '--network', '-n', '-in',
        type=_text_type,
        default=None,
        help='Path to the model network file of the external tool (e.g caffe prototxt, keras json')

    parser.add_argument(
        '--dstPath', '-d', '-o',
        type=_text_type,
        required=True,
        help='Path to save the IR model.')

    parser.add_argument(
        '--inNodeName', '-inode',
        nargs='+',
        type=_text_type,
        default=None,
        help="[Tensorflow] Input nodes' name of the graph.")

    parser.add_argument(
        '--dstNodeName', '-node',
        nargs='+',
        type=_text_type,
        default=None,
        help="[Tensorflow] Output nodes' name of the graph.")

    parser.add_argument(
        '--inputShape',
        nargs='+',
        type=_text_type,
        default=None,
        help='[Tensorflow/MXNet/Caffe2/Torch7] Input shape of model (channel, height, width)')


    # Caffe
    parser.add_argument(
        '--caffePhase',
        type=_text_type,
        default='TRAIN',
        help='[Caffe] Convert the specific phase of caffe model.')


    # Darknet
    parser.add_argument(
        '--darknetStart',
        type=_text_type,
        choices=["0", "1"],
        help='[Darknet] Parse the darkent model weight file from the start.')

    return parser


def _main():
    parser = _get_parser()
    args = parser.parse_args()
    ret = _convert(args)
    _sys.exit(int(ret)) # cast to int or else the exit code is always 1


if __name__ == '__main__':
    _main()



from collections import namedtuple
import math

TensorShape = namedtuple('TensorShape', ['batch_size', 'channels', 'height', 'width'])


def get_kernel_extents(params, dilation):
    ko_h = dilation * (int(params.k_h) - 1) + 1
    ko_w = dilation * (int(params.k_w) - 1) + 1
    return ko_h, ko_w

def get_filter_output_shape(i_h, i_w, dilation, params, round_func):
    ko_h, ko_w = get_kernel_extents(params, dilation)

    o_h = (i_h + 2 * params.p_h - ko_h) / float(params.s_h) + 1
    o_w = (i_w + 2 * params.p_w - ko_w) / float(params.s_w) + 1
    return (int(round_func(o_h)), int(round_func(o_w)))


def get_strided_kernel_output_shape(node, round_func):
    assert node.layer is not None
    input_shape = node.get_only_parent()[0].output_shape
    params = node.kernel_parameters
    dilation = node.parameters.dilation[0] if hasattr(node.parameters, 'dilation') and node.parameters.dilation else 1

    o_h, o_w = get_filter_output_shape(input_shape.height, input_shape.width,
                                       dilation, params, round_func)
    params = node.parameters
    has_c_o = hasattr(params, 'num_output')
    c = params.num_output if has_c_o else input_shape.channels
    return TensorShape(input_shape.batch_size, c, o_h, o_w)


def shape_not_implemented(node):
    raise NotImplementedError

def shape_deconvolution(node):
    input_shape = node.get_only_parent()[0].output_shape
    params = node.kernel_parameters
    dilation = 1 if len(node.parameters.dilation) == 0 else node.parameters.dilation[0]

    ko_h, ko_w = get_kernel_extents(params, dilation)
    o_h = int(params.s_h) * (input_shape.height - 1) + ko_h - 2 * int(params.p_h)
    o_w = int(params.s_w) * (input_shape.width - 1) + ko_w - 2 * int(params.p_w)

    has_c_o = hasattr(node.parameters, 'num_output')
    c = node.parameters.num_output if has_c_o else input_shape.channels
    return TensorShape(input_shape.batch_size, c, o_h, o_w)

def shape_identity(node):
    assert len(node.parents) > 0
    return node.parents[0][0].output_shape


def shape_scalar(node):
    return TensorShape(1, 1, 1, 1)

def shape_reshape(node):
    last_shape = node.get_only_parent()[0].output_shape
    shapes = []
    for idx, shape in enumerate(node.layer.reshape_param.shape.dim):
        shapes.append(shape if shape != 0 else last_shape[idx])
    return TensorShape(shapes[0], shapes[1], shapes[2], shapes[3])

def shape_data(node):
    if node.output_shape:
        # Old-style input specification
        return node.output_shape
    try:
        # New-style input specification
        return tuple(map(int, node.parameters.shape[0].dim))
    except:
        # We most likely have a data layer on our hands. The problem is,
        # Caffe infers the dimensions of the data from the source (eg: LMDB).
        # We want to avoid reading datasets here. Fail for now.
        # This can be temporarily fixed by transforming the data layer to
        # Caffe's "input" layer (as is usually used in the "deploy" version).
        # TODO: Find a better solution for this.
        pass


def shape_mem_data(node):
    params = node.parameters
    return TensorShape(params.batch_size, params.channels, params.height, params.width)


def shape_concat(node):
    axis = node.parameters.axis
    output_shape = None
    for parent, idx in node.parents:
        if output_shape is None:
            output_shape = list(parent.output_shape)
        else:
            output_shape[axis] += parent.output_shape[axis]
    return tuple(output_shape)


def shape_convolution(node):
    return get_strided_kernel_output_shape(node, math.floor)


def shape_pool(node):
    if node.parameters.global_pooling:
        return shape_global_pooling(node)
    return get_strided_kernel_output_shape(node, math.ceil)


def shape_inner_product(node):
    input_shape = node.get_only_parent()[0].output_shape
    return TensorShape(input_shape.batch_size, node.parameters.num_output, 1, 1)


def shape_global_pooling(node):
    input_shape = node.get_only_parent()[0].output_shape
    params = node.kernel_parameters
    has_c_o = hasattr(params, 'num_output')
    c = params.num_output if has_c_o else input_shape.channels
    return TensorShape(input_shape.batch_size, c, 1, 1)  # Output height and width is 1 when global_pooling


def shape_split(node):
    input_shape = node.get_only_parent()[0].output_shape
    return TensorShape(input_shape.batch_size, input_shape.channels, input_shape.height, input_shape.width)


def shape_flatten(node):
    input_shape = node.get_only_parent()[0].output_shape
    return TensorShape(input_shape.batch_size, input_shape.channels * input_shape.height * input_shape.width, 1, 1)




#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import numpy as np
import sys
import os
import tensorflow as tf
from onnx_tf.backend import prepare
from mmdnn.conversion.examples.imagenet_test import TestKit

class TestONNX(TestKit):

    def __init__(self):
        super(TestONNX, self).__init__()
        self.model = prepare(self.MainModel.KitModel(self.args.w))
        # self.input, self.model, self.testop = self.MainModel.KitModel(self.args.w)


    def preprocess(self, image_path):
        x = super(TestONNX, self).preprocess(image_path)
        self.data = np.expand_dims(x, 0)


    def print_result(self):
        predict = self.model.run(self.data)[0]
        super(TestONNX, self).print_result(predict)


    def print_intermediate_result(self, layer_name, if_transpose = False):
        # testop = tf.get_default_graph().get_operation_by_name(layer_name)
        testop = self.testop
        with tf.Session() as sess:
            init = tf.global_variables_initializer()
            sess.run(init)
            intermediate_output = sess.run(testop, feed_dict = {self.input : self.data})

        super(TestONNX, self).print_intermediate_result(intermediate_output, if_transpose)


    def inference(self, image_path):
        self.preprocess(image_path)

        # self.print_intermediate_result('conv1_7x7_s2_1', True)

        self.print_result()

        self.test_truth()

if __name__=='__main__':
    tester = TestONNX()
    if tester.args.dump:
        tester.dump()
    else:
        tester.inference(tester.args.image)



import sys

SHARED_CAFFE_RESOLVER = None


class CaffeResolver(object):
    def __init__(self):
        self.import_caffe()

    def import_caffe(self):
        self.caffe = None

        # try:
        import caffe
        self.caffe = caffe
        # except ImportError:
        #     # # Fall back to the protobuf implementation
        #     # from mmdnn.conversion.caffe import caffe_pb2
        #     # self.caffepb = caffe_pb2
        #     # show_fallback_warning()
        if self.caffe:
            self.caffepb = self.caffe.proto.caffe_pb2
        self.NetParameter = self.caffepb.NetParameter

    def has_pycaffe(self):
        return self.caffe is not None


def get_caffe_resolver():
    global SHARED_CAFFE_RESOLVER
    if SHARED_CAFFE_RESOLVER is None:
        SHARED_CAFFE_RESOLVER = CaffeResolver()
    return SHARED_CAFFE_RESOLVER


def has_pycaffe():
    return get_caffe_resolver().has_pycaffe()


def show_fallback_warning():
    msg = '''
------------------------------------------------------------
    WARNING: PyCaffe not found!
    Falling back to a pure protocol buffer implementation.
    * Conversions will be drastically slower.
    * This backend is UNTESTED!
------------------------------------------------------------

'''
    sys.stderr.write(msg)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import numpy as np
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType


class Parser(object):

    def __init__(self):
        self.IR_graph = GraphDef()
        self.weight_loaded = False

        # name --> (weight_name --> ndarray)
        self.weights = dict()


    def run(self, dest_path):
        self.gen_IR()
        self.save_to_json(dest_path + ".json")
        self.save_to_proto(dest_path + ".pb")
        self.save_weights(dest_path + ".npy")


    @property
    def src_graph(self):
        raise NotImplementedError


    def get_son(self, name, path, set_flag = False):
        return self.src_graph.get_son(name, path, set_flag)


    def get_parent(self, name, path, set_flag = False):
        return self.src_graph.get_parent(name, path, set_flag)


    def set_weight(self, layer_name, weight_name, data):
        if not layer_name in self.weights:
            self.weights[layer_name] = dict()
        layer = self.weights[layer_name]
        layer[weight_name] = data


    def save_to_json(self, filename):
        import google.protobuf.json_format as json_format
        json_str = json_format.MessageToJson(self.IR_graph, preserving_proto_field_name = True)

        with open(filename, "w") as of:
            of.write(json_str)

        print ("IR network structure is saved as [{}].".format(filename))

        return json_str


    def save_to_proto(self, filename):
        proto_str = self.IR_graph.SerializeToString()
        with open(filename, 'wb') as of:
            of.write(proto_str)

        print ("IR network structure is saved as [{}].".format(filename))

        return proto_str


    def save_weights(self, filename):
        if self.weight_loaded:
            with open(filename, 'wb') as of:
                np.save(of, self.weights)
            print ("IR weights are saved as [{}].".format(filename))

        else:
            print ("Warning: weights are not loaded.")


    def convert_inedge(self, source_node, IR_node, start_idx = 0, end_idx = None):
        if end_idx == None: end_idx = len(source_node.in_edges)
        for idx in range(start_idx, end_idx):
            IR_node.input.append(self.src_graph.get_node(source_node.in_edges[idx]).real_name.lstrip('_'))


    @staticmethod
    def channel_first_conv_kernel_to_IR(tensor):
        dim = tensor.ndim
        tensor = np.transpose(tensor, list(range(2, dim)) + [1, 0])
        return tensor


    @staticmethod
    def channel_first_shape_to_IR(shape):
        return [shape[0]] + list(shape[2:]) + [shape[1]]

    @staticmethod
    def channel_first_axis_to_IR(index):
        if index == 0:
            return 0
        elif index == 1:
            return -1
        else:
            return index - 1



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------
import os

import coremltools
from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph


class CoremlGraphNode(GraphNode):

    def __init__(self, layer):
        super(CoremlGraphNode, self).__init__(layer)


    @property
    def name(self):
        return self.layer.name


    @property
    def type(self):
        return self.layer.__class__.__name__


    @property
    def coreml_layer(self):
        return self.layer



class CoremlGraph(Graph):

    def __init__(self, model):
        from coremltools.proto import Model_pb2

        # sanity check.
        if not isinstance(model, Model_pb2.Model):
            raise TypeError("Coreml layer of type %s is not supported." % type(model))
        super(CoremlGraph, self).__init__(model)
        self.model = model


    def build(self):
        self.input_layers = list()

        # input layer

        for layer in self.model.description.input:
            self.layer_map[layer.name] = CoremlGraphNode(layer)
            self.layer_name_map[layer.name] = layer.name


        model_type = self.model.WhichOneof('Type')
        if model_type == 'neuralNetworkClassifier':
            # build each layer
            for layer in self.model.neuralNetworkClassifier.layers:
                self.layer_map[layer.name] = CoremlGraphNode(layer)
                self.layer_name_map[layer.name] = layer.name

            # if A.output == B.input, then make the connection: A -> B
            for layerA in self.model.neuralNetworkClassifier.layers:
                for layerB in self.model.neuralNetworkClassifier.layers:
                    for A in layerA.output:
                        for B in layerB.input:
                            if A == B :
                                # print('{0:20}->     {1:20}'.format(layerA.name, layerB.name))
                                self._make_connection(layerA.name, layerB.name)

            # if A.name == B.input, then make the connection: A -> B, here A is the input
            for layerA in self.model.description.input:
                for layerB in self.model.neuralNetworkClassifier.layers:
                    for B in layerB.input:
                        if layerA.name == B:
                            self._make_connection(layerA.name, layerB.name)
        elif model_type == 'neuralNetwork':
            # build each layer
            for layer in self.model.neuralNetwork.layers:
                self.layer_map[layer.name] = CoremlGraphNode(layer)
                self.layer_name_map[layer.name] = layer.name

            # if A.output == B.input, then make the connection: A -> B
            for layerA in self.model.neuralNetwork.layers:
                for layerB in self.model.neuralNetwork.layers:
                    for A in layerA.output:
                        for B in layerB.input:
                            if A == B :
                                # print('{0:20}->     {1:20}'.format(layerA.name, layerB.name))
                                self._make_connection(layerA.name, layerB.name)
            # if A.name == B.input, then make the connection: A -> B, here A is the input
            for layerA in self.model.description.input:
                for layerB in self.model.neuralNetwork.layers:
                    for B in layerB.input:
                        if layerA.name == B:
                            self._make_connection(layerA.name, layerB.name)
        elif model_type == 'neuralNetworkRegressor':
            # build each layer
            for layer in self.model.neuralNetworkRegressor.layers:
                self.layer_map[layer.name] = CoremlGraphNode(layer)
                self.layer_name_map[layer.name] = layer.name

            # if A.output == B.input, then make the connection: A -> B
            for layerA in self.model.neuralNetworkRegressor.layers:
                for layerB in self.model.neuralNetworkRegressor.layers:
                    for A in layerA.output:
                        for B in layerB.input:
                            if A == B :
                                # print('{0:20}->     {1:20}'.format(layerA.name, layerB.name))
                                self._make_connection(layerA.name, layerB.name)
            # if A.name == B.input, then make the connection: A -> B, here A is the input
            for layerA in self.model.description.input:
                for layerB in self.model.neuralNetworkRegressor.layers:
                    for B in layerB.input:
                        if layerA.name == B:
                            self._make_connection(layerA.name, layerB.name)
        else:
            assert False



            # The information of the layer
        super(CoremlGraph, self).build()






# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graph.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='graph.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n\x0bgraph.proto\"3\n\x08GraphDef\x12\x16\n\x04node\x18\x01 \x03(\x0b\x32\x08.NodeDef\x12\x0f\n\x07version\x18\x02 \x01(\x05\"\x8d\x01\n\x07NodeDef\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02op\x18\x02 \x01(\t\x12\r\n\x05input\x18\x03 \x03(\t\x12 \n\x04\x61ttr\x18\x04 \x03(\x0b\x32\x12.NodeDef.AttrEntry\x1a\x37\n\tAttrEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x19\n\x05value\x18\x02 \x01(\x0b\x32\n.AttrValue:\x02\x38\x01\"\xea\x02\n\tAttrValue\x12$\n\x04list\x18\x01 \x01(\x0b\x32\x14.AttrValue.ListValueH\x00\x12\x0b\n\x01s\x18\x02 \x01(\x0cH\x00\x12\x0b\n\x01i\x18\x03 \x01(\x03H\x00\x12\x0b\n\x01\x66\x18\x04 \x01(\x02H\x00\x12\x0b\n\x01\x62\x18\x05 \x01(\x08H\x00\x12\x19\n\x04type\x18\x06 \x01(\x0e\x32\t.DataTypeH\x00\x12\x1d\n\x05shape\x18\x07 \x01(\x0b\x32\x0c.TensorShapeH\x00\x12 \n\x06tensor\x18\x08 \x01(\x0b\x32\x0e.LiteralTensorH\x00\x1a\x9d\x01\n\tListValue\x12\t\n\x01s\x18\x02 \x03(\x0c\x12\r\n\x01i\x18\x03 \x03(\x03\x42\x02\x10\x01\x12\r\n\x01\x66\x18\x04 \x03(\x02\x42\x02\x10\x01\x12\r\n\x01\x62\x18\x05 \x03(\x08\x42\x02\x10\x01\x12\x1b\n\x04type\x18\x06 \x03(\x0e\x32\t.DataTypeB\x02\x10\x01\x12\x1b\n\x05shape\x18\x07 \x03(\x0b\x32\x0c.TensorShape\x12\x1e\n\x06tensor\x18\x08 \x03(\x0b\x32\x0e.LiteralTensorB\x07\n\x05value\"e\n\x0bTensorShape\x12\x1d\n\x03\x64im\x18\x02 \x03(\x0b\x32\x10.TensorShape.Dim\x12\x14\n\x0cunknown_rank\x18\x03 \x01(\x08\x1a!\n\x03\x44im\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\"\xb0\x02\n\rLiteralTensor\x12\x18\n\x05\x64type\x18\x01 \x01(\x0e\x32\t.DataType\x12\"\n\x0ctensor_shape\x18\x02 \x01(\x0b\x32\x0c.TensorShape\x12\x16\n\x0eversion_number\x18\x03 \x01(\x05\x12\x16\n\x0etensor_content\x18\x04 \x01(\x0c\x12\x13\n\x07int_val\x18\x05 \x03(\x05\x42\x02\x10\x01\x12\x14\n\x08uint_val\x18\x06 \x03(\x05\x42\x02\x10\x01\x12\x15\n\tint64_val\x18\x07 \x03(\x03\x42\x02\x10\x01\x12\x16\n\nuint64_val\x18\x08 \x03(\x03\x42\x02\x10\x01\x12\x15\n\tfloat_val\x18\t \x03(\x02\x42\x02\x10\x01\x12\x16\n\ndouble_val\x18\n \x03(\x01\x42\x02\x10\x01\x12\x14\n\x08\x62ool_val\x18\x0b \x03(\x08\x42\x02\x10\x01\x12\x12\n\nstring_val\x18\x0c \x03(\x0c*\xff\x01\n\x08\x44\x61taType\x12\x10\n\x0c\x44T_UNDEFINED\x10\x00\x12\x0b\n\x07\x44T_INT8\x10\x01\x12\x0c\n\x08\x44T_INT16\x10\x02\x12\x0c\n\x08\x44T_INT32\x10\x03\x12\x0c\n\x08\x44T_INT64\x10\x04\x12\x0c\n\x08\x44T_UINT8\x10\x05\x12\r\n\tDT_UINT16\x10\x06\x12\r\n\tDT_UINT32\x10\x07\x12\r\n\tDT_UINT64\x10\x08\x12\x0e\n\nDT_FLOAT16\x10\t\x12\x0e\n\nDT_FLOAT32\x10\n\x12\x0e\n\nDT_FLOAT64\x10\x0b\x12\x10\n\x0c\x44T_COMPLEX64\x10\x0c\x12\x11\n\rDT_COMPLEX128\x10\r\x12\x0b\n\x07\x44T_BOOL\x10\x0e\x12\r\n\tDT_STRING\x10\x0f\x62\x06proto3')
)

_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='DataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DT_UNDEFINED', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT8', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT16', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT32', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT64', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_UINT8', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_UINT16', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_UINT32', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_UINT64', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_FLOAT16', index=9, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_FLOAT32', index=10, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_FLOAT64', index=11, number=11,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_COMPLEX64', index=12, number=12,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_COMPLEX128', index=13, number=13,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_BOOL', index=14, number=14,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_STRING', index=15, number=15,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=988,
  serialized_end=1243,
)
_sym_db.RegisterEnumDescriptor(_DATATYPE)

DataType = enum_type_wrapper.EnumTypeWrapper(_DATATYPE)
DT_UNDEFINED = 0
DT_INT8 = 1
DT_INT16 = 2
DT_INT32 = 3
DT_INT64 = 4
DT_UINT8 = 5
DT_UINT16 = 6
DT_UINT32 = 7
DT_UINT64 = 8
DT_FLOAT16 = 9
DT_FLOAT32 = 10
DT_FLOAT64 = 11
DT_COMPLEX64 = 12
DT_COMPLEX128 = 13
DT_BOOL = 14
DT_STRING = 15



_GRAPHDEF = _descriptor.Descriptor(
  name='GraphDef',
  full_name='GraphDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node', full_name='GraphDef.node', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='GraphDef.version', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15,
  serialized_end=66,
)


_NODEDEF_ATTRENTRY = _descriptor.Descriptor(
  name='AttrEntry',
  full_name='NodeDef.AttrEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='NodeDef.AttrEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='NodeDef.AttrEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=210,
)

_NODEDEF = _descriptor.Descriptor(
  name='NodeDef',
  full_name='NodeDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='NodeDef.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='op', full_name='NodeDef.op', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='input', full_name='NodeDef.input', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attr', full_name='NodeDef.attr', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_NODEDEF_ATTRENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=69,
  serialized_end=210,
)


_ATTRVALUE_LISTVALUE = _descriptor.Descriptor(
  name='ListValue',
  full_name='AttrValue.ListValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='s', full_name='AttrValue.ListValue.s', index=0,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='i', full_name='AttrValue.ListValue.i', index=1,
      number=3, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='f', full_name='AttrValue.ListValue.f', index=2,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='b', full_name='AttrValue.ListValue.b', index=3,
      number=5, type=8, cpp_type=7, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='type', full_name='AttrValue.ListValue.type', index=4,
      number=6, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='shape', full_name='AttrValue.ListValue.shape', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tensor', full_name='AttrValue.ListValue.tensor', index=6,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=409,
  serialized_end=566,
)

_ATTRVALUE = _descriptor.Descriptor(
  name='AttrValue',
  full_name='AttrValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='list', full_name='AttrValue.list', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='s', full_name='AttrValue.s', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='i', full_name='AttrValue.i', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='f', full_name='AttrValue.f', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='b', full_name='AttrValue.b', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='AttrValue.type', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shape', full_name='AttrValue.shape', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tensor', full_name='AttrValue.tensor', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ATTRVALUE_LISTVALUE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='AttrValue.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=213,
  serialized_end=575,
)


_TENSORSHAPE_DIM = _descriptor.Descriptor(
  name='Dim',
  full_name='TensorShape.Dim',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='size', full_name='TensorShape.Dim.size', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='TensorShape.Dim.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=645,
  serialized_end=678,
)

_TENSORSHAPE = _descriptor.Descriptor(
  name='TensorShape',
  full_name='TensorShape',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='TensorShape.dim', index=0,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unknown_rank', full_name='TensorShape.unknown_rank', index=1,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TENSORSHAPE_DIM, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=577,
  serialized_end=678,
)


_LITERALTENSOR = _descriptor.Descriptor(
  name='LiteralTensor',
  full_name='LiteralTensor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dtype', full_name='LiteralTensor.dtype', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tensor_shape', full_name='LiteralTensor.tensor_shape', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version_number', full_name='LiteralTensor.version_number', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tensor_content', full_name='LiteralTensor.tensor_content', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int_val', full_name='LiteralTensor.int_val', index=4,
      number=5, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='uint_val', full_name='LiteralTensor.uint_val', index=5,
      number=6, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='int64_val', full_name='LiteralTensor.int64_val', index=6,
      number=7, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='uint64_val', full_name='LiteralTensor.uint64_val', index=7,
      number=8, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='float_val', full_name='LiteralTensor.float_val', index=8,
      number=9, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='double_val', full_name='LiteralTensor.double_val', index=9,
      number=10, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='bool_val', full_name='LiteralTensor.bool_val', index=10,
      number=11, type=8, cpp_type=7, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='string_val', full_name='LiteralTensor.string_val', index=11,
      number=12, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=681,
  serialized_end=985,
)

_GRAPHDEF.fields_by_name['node'].message_type = _NODEDEF
_NODEDEF_ATTRENTRY.fields_by_name['value'].message_type = _ATTRVALUE
_NODEDEF_ATTRENTRY.containing_type = _NODEDEF
_NODEDEF.fields_by_name['attr'].message_type = _NODEDEF_ATTRENTRY
_ATTRVALUE_LISTVALUE.fields_by_name['type'].enum_type = _DATATYPE
_ATTRVALUE_LISTVALUE.fields_by_name['shape'].message_type = _TENSORSHAPE
_ATTRVALUE_LISTVALUE.fields_by_name['tensor'].message_type = _LITERALTENSOR
_ATTRVALUE_LISTVALUE.containing_type = _ATTRVALUE
_ATTRVALUE.fields_by_name['list'].message_type = _ATTRVALUE_LISTVALUE
_ATTRVALUE.fields_by_name['type'].enum_type = _DATATYPE
_ATTRVALUE.fields_by_name['shape'].message_type = _TENSORSHAPE
_ATTRVALUE.fields_by_name['tensor'].message_type = _LITERALTENSOR
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['list'])
_ATTRVALUE.fields_by_name['list'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['s'])
_ATTRVALUE.fields_by_name['s'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['i'])
_ATTRVALUE.fields_by_name['i'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['f'])
_ATTRVALUE.fields_by_name['f'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['b'])
_ATTRVALUE.fields_by_name['b'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['type'])
_ATTRVALUE.fields_by_name['type'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['shape'])
_ATTRVALUE.fields_by_name['shape'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_ATTRVALUE.oneofs_by_name['value'].fields.append(
  _ATTRVALUE.fields_by_name['tensor'])
_ATTRVALUE.fields_by_name['tensor'].containing_oneof = _ATTRVALUE.oneofs_by_name['value']
_TENSORSHAPE_DIM.containing_type = _TENSORSHAPE
_TENSORSHAPE.fields_by_name['dim'].message_type = _TENSORSHAPE_DIM
_LITERALTENSOR.fields_by_name['dtype'].enum_type = _DATATYPE
_LITERALTENSOR.fields_by_name['tensor_shape'].message_type = _TENSORSHAPE
DESCRIPTOR.message_types_by_name['GraphDef'] = _GRAPHDEF
DESCRIPTOR.message_types_by_name['NodeDef'] = _NODEDEF
DESCRIPTOR.message_types_by_name['AttrValue'] = _ATTRVALUE
DESCRIPTOR.message_types_by_name['TensorShape'] = _TENSORSHAPE
DESCRIPTOR.message_types_by_name['LiteralTensor'] = _LITERALTENSOR
DESCRIPTOR.enum_types_by_name['DataType'] = _DATATYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GraphDef = _reflection.GeneratedProtocolMessageType('GraphDef', (_message.Message,), dict(
  DESCRIPTOR = _GRAPHDEF,
  __module__ = 'graph_pb2'
  # @@protoc_insertion_point(class_scope:GraphDef)
  ))
_sym_db.RegisterMessage(GraphDef)

NodeDef = _reflection.GeneratedProtocolMessageType('NodeDef', (_message.Message,), dict(

  AttrEntry = _reflection.GeneratedProtocolMessageType('AttrEntry', (_message.Message,), dict(
    DESCRIPTOR = _NODEDEF_ATTRENTRY,
    __module__ = 'graph_pb2'
    # @@protoc_insertion_point(class_scope:NodeDef.AttrEntry)
    ))
  ,
  DESCRIPTOR = _NODEDEF,
  __module__ = 'graph_pb2'
  # @@protoc_insertion_point(class_scope:NodeDef)
  ))
_sym_db.RegisterMessage(NodeDef)
_sym_db.RegisterMessage(NodeDef.AttrEntry)

AttrValue = _reflection.GeneratedProtocolMessageType('AttrValue', (_message.Message,), dict(

  ListValue = _reflection.GeneratedProtocolMessageType('ListValue', (_message.Message,), dict(
    DESCRIPTOR = _ATTRVALUE_LISTVALUE,
    __module__ = 'graph_pb2'
    # @@protoc_insertion_point(class_scope:AttrValue.ListValue)
    ))
  ,
  DESCRIPTOR = _ATTRVALUE,
  __module__ = 'graph_pb2'
  # @@protoc_insertion_point(class_scope:AttrValue)
  ))
_sym_db.RegisterMessage(AttrValue)
_sym_db.RegisterMessage(AttrValue.ListValue)

TensorShape = _reflection.GeneratedProtocolMessageType('TensorShape', (_message.Message,), dict(

  Dim = _reflection.GeneratedProtocolMessageType('Dim', (_message.Message,), dict(
    DESCRIPTOR = _TENSORSHAPE_DIM,
    __module__ = 'graph_pb2'
    # @@protoc_insertion_point(class_scope:TensorShape.Dim)
    ))
  ,
  DESCRIPTOR = _TENSORSHAPE,
  __module__ = 'graph_pb2'
  # @@protoc_insertion_point(class_scope:TensorShape)
  ))
_sym_db.RegisterMessage(TensorShape)
_sym_db.RegisterMessage(TensorShape.Dim)

LiteralTensor = _reflection.GeneratedProtocolMessageType('LiteralTensor', (_message.Message,), dict(
  DESCRIPTOR = _LITERALTENSOR,
  __module__ = 'graph_pb2'
  # @@protoc_insertion_point(class_scope:LiteralTensor)
  ))
_sym_db.RegisterMessage(LiteralTensor)


_NODEDEF_ATTRENTRY.has_options = True
_NODEDEF_ATTRENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
_ATTRVALUE_LISTVALUE.fields_by_name['i'].has_options = True
_ATTRVALUE_LISTVALUE.fields_by_name['i']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_ATTRVALUE_LISTVALUE.fields_by_name['f'].has_options = True
_ATTRVALUE_LISTVALUE.fields_by_name['f']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_ATTRVALUE_LISTVALUE.fields_by_name['b'].has_options = True
_ATTRVALUE_LISTVALUE.fields_by_name['b']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_ATTRVALUE_LISTVALUE.fields_by_name['type'].has_options = True
_ATTRVALUE_LISTVALUE.fields_by_name['type']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['int_val'].has_options = True
_LITERALTENSOR.fields_by_name['int_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['uint_val'].has_options = True
_LITERALTENSOR.fields_by_name['uint_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['int64_val'].has_options = True
_LITERALTENSOR.fields_by_name['int64_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['uint64_val'].has_options = True
_LITERALTENSOR.fields_by_name['uint64_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['float_val'].has_options = True
_LITERALTENSOR.fields_by_name['float_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['double_val'].has_options = True
_LITERALTENSOR.fields_by_name['double_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_LITERALTENSOR.fields_by_name['bool_val'].has_options = True
_LITERALTENSOR.fields_by_name['bool_val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
# @@protoc_insertion_point(module_scope)



import numpy as np

DEFAULT_PADDING = 'SAME'


def layer(op):
    '''Decorator for composable network layers.'''

    def layer_decorated(self, *args, **kwargs):
        # Automatically set a name if not provided.
        name = kwargs.setdefault('name', self.get_unique_name(op.__name__))
        # Figure out the layer inputs.
        assert len(args) >= 1
        if len(args) == 1:
            layer_inputs = args[0]
        else:
            layer_inputs = list(args) 
        layer_output = op(self, layer_inputs, **kwargs)
        # print('op: %s   shape: %s' % (op, layer_output._keras_shape))
        # print('op: %s   shape: %s' % (op, layer_output.get_shape().as_list()))
        # Add to layer LUT.
        self.layers[name] = layer_output
        self.output = layer_output
        return layer_output

    return layer_decorated


class Network(object):

    def __init__(self, trainable=False):
        self.output = None
        self.layers = {}
        self.trainable = trainable
        self.setup()

    def setup(self):
        raise NotImplementedError('Must be implemented by the subclass')

    def load(self, data_path, session, ignore_missing=False):
        raise NotImplementedError('Must be implemented by the subclass')
    
    def input(self, shape, name):
        raise NotImplementedError('Must be implemented by the subclass')

    def get_output(self):
        raise NotImplementedError('Must be implemented by the subclass')

    def get_unique_name(self, prefix):
        raise NotImplementedError('Must be implemented by the subclass')
    
    @layer
    def conv(self, input, k_h, k_w, c_o, s_h, s_w, p_h, p_w, name, group=1, biased=True):
        raise NotImplementedError('Must be implemented by the subclass')
    
    @layer
    def deconv(self, input, c_o, k_h, k_w, s_h, s_w, p_h, p_w, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def relu(self, input, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def sigmoid(self, input, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def max_pool(self, input, k_h, k_w, s_h, s_w, p_h, p_w, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def avg_pool(self, input, k_h, k_w, s_h, s_w, p_h, p_w, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def lrn(self, input, local_size, alpha, beta, name, bias=1):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def concat(self, inputs, axis, name):
        raise NotImplementedError('Must be implemented by the subclass')
    
    @layer
    def add(self, inputs, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def fc(self, input, num_out, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def softmax(self, input, name):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def batch_normalization(self, input, name, epsilon=0.00001, scale_offset=True):
        raise NotImplementedError('Must be implemented by the subclass')

    @layer
    def dropout(self, input, keep_prob, name):
        raise NotImplementedError('Must be implemented by the subclass')
    
    @layer
    def crop(self, inputs, offset, name):
        raise NotImplementedError('Must be implemented by the subclass')


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import numpy as np
import sys
import os
from six import text_type as _text_type
from mmdnn.conversion.examples.imagenet_test import TestKit
import caffe


class TestCaffe(TestKit):

    def __init__(self):
        super(TestCaffe, self).__init__()

        self.truth['caffe']['alexnet'] = [(657, 0.41121054), (744, 0.20789708), (847, 0.086725503), (821, 0.05908291), (595, 0.058017164)]

        if self.args.dump:
            self.dump_net = self.args.dump + '.prototxt'
            self.dump_weight = self.args.dump + '.caffemodel'
        else:
            self.dump_net = 'tmp.prototxt'
            self.dump_weight = 'tmp.caffemodel'

        self.MainModel.make_net(self.dump_net)
        self.MainModel.gen_weight(self.args.w, self.dump_weight, self.dump_net)
        self.model = caffe.Net(self.dump_net, self.dump_weight, caffe.TEST)

    def preprocess(self, image_path):
        x = super(TestCaffe, self).preprocess(image_path)
        # caffe uses NCHW
        x = np.transpose(x, [2, 0, 1])
        self.data = np.expand_dims(x, 0)


    def print_result(self):
        self.model.blobs['input'].data[...] = self.data
        predict = self.model.forward()[self.model._layer_names[-1]][0]
        super(TestCaffe, self).print_result(predict)


    def print_intermediate_result(self, layer_name, if_transpose = False):
        intermediate_output = self.model.blobs[layer_name].data[0]
        super(TestCaffe, self).print_intermediate_result(intermediate_output, if_transpose)


    def inference(self, image_path):
        self.preprocess(image_path)

        self.print_result()

        # self.print_intermediate_result('pooling0', False)

        self.test_truth()

        # delete tmp model files
        if os.path.isfile(self.dump_net):
            os.remove(self.dump_net)
        if os.path.isfile(self.dump_weight):
            os.remove(self.dump_weight)


    def dump(self):
        print ('Caffe model files are saved as [{}] and [{}], generated by [{}.py] and [{}].'.format(
            self.dump_net, self.dump_weight, self.args.n, self.args.w))


if __name__=='__main__':
    tester = TestCaffe()
    if tester.args.dump:
        tester.dump()
    else:
        tester.inference(tester.args.image)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
from six import text_type as _text_type
import keras
from mmdnn.conversion.examples.imagenet_test import TestKit

networks_map = {
    'inception_v3'      : lambda : keras.applications.inception_v3.InceptionV3(input_shape=(299, 299, 3)),
    'vgg16'             : lambda : keras.applications.vgg16.VGG16(),
    'vgg19'             : lambda : keras.applications.vgg19.VGG19(),
    'resnet'            : lambda : keras.applications.resnet50.ResNet50(),
    'mobilenet'         : lambda : keras.applications.mobilenet.MobileNet(),
    'xception'          : lambda : keras.applications.xception.Xception(input_shape=(299, 299, 3)),
    'inception_resnet'  : lambda : keras.applications.inception_resnet_v2.InceptionResNetV2()
}

image_size = {
    'inception_v3'      : 299,
    'vgg16'             : 224,
    'vgg19'             : 224,
    'resnet'            : 224,
    'mobilenet'         : 224,
    'xception'          : 299,
    'inception_resnet'  : 299
}

def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network',
                        type=_text_type, help='Model Type', required=True,
                        choices=networks_map.keys())

    parser.add_argument('-i', '--image',
                        type=_text_type, help='Test Image Path')

    args = parser.parse_args()

    model = networks_map.get(args.network)
    if model is None:
        raise NotImplementedError("Unknown keras application [{}]".format(args.network))

    model = model()
    # save network structure as JSON
    json_string = model.to_json()
    with open("imagenet_{}.json".format(args.network), "w") as of:
        of.write(json_string)

    print("Network structure is saved as [imagenet_{}.json].".format(args.network))

    model.save_weights('imagenet_{}.h5'.format(args.network))

    print("Network weights are saved as [imagenet_{}.h5].".format(args.network))

    if args.image:
        import numpy as np
        func = TestKit.preprocess_func['keras'][args.network]
        img = func(args.image)
        img = np.expand_dims(img, axis=0)
        predict = model.predict(img)
        predict = np.squeeze(predict)
        top_indices = predict.argsort()[-5:][::-1]
        result = [(i, predict[i]) for i in top_indices]
        print(result)

        # layer_name = 'block2_pool'
        # intermediate_layer_model = keras.Model(inputs=model.input,
        #                                  outputs=model.get_layer(layer_name).output)
        # intermediate_output = intermediate_layer_model.predict(img)
        # print (intermediate_output)
        # print (intermediate_output.shape)
        # print ("%.30f" % np.sum(intermediate_output))


if __name__ == '__main__':
    _main()



import torch


def save_model(MainModel, network_filepath, weight_filepath, dump_filepath):
    model = MainModel.KitModel(weight_filepath)
    model.eval()
    torch.save(model, dump_filepath)
    print('PyTorch model file is saved as [{}], generated by [{}.py] and [{}]. Notice that you may need [{}.py] to load the model back.'.format(
        dump_filepath, network_filepath, weight_filepath, network_filepath))



from mmdnn.conversion.rewriter.graph_matcher import *
from mmdnn.conversion.tensorflow.tensorflow_graph import *
import numpy as np


'''batch size pattern in tensorflow. Note: do not include _number in name'''
static_rnn_batch_size_pattern = OpTypePattern('ExpandDims', name='static_rnn_batch_size', inputs=[
    OpTypePattern('StridedSlice', inputs=[
        OpTypePattern('Shape', inputs=[
            OpTypePattern('*', name='input')
        ]),
        OpTypePattern('Const'),
        OpTypePattern('Const'),
        OpTypePattern('Const')
    ]),
    OpTypePattern('Const')
])

'''rnn h zero pattern in tensorflow.'''
static_rnn_h_zero_pattern = OpTypePattern('Fill', name='h_zero', inputs=[
    OpTypePattern('ConcatV2|Concat', inputs=[
        OpTypePattern('*', name='input'),
        OpTypePattern('Const', name='fill_size'),
        OpTypePattern('Const')
    ]),
    OpTypePattern('Const', name='fill_value')
])

''''split pattern in gru cell in tensorflow'''
gru_xc_pattern = OpTypePattern('Split', name='xc', inputs=[
    OpTypePattern("Const"), # axis for split
    OpTypePattern("Sigmoid", inputs=[
        OpTypePattern("BiasAdd", name="bias_add", inputs=[
            OpTypePattern("MatMul", inputs=[
                OpTypePattern("ConcatV2|Concat", name="xh"),
                OpTypePattern("Identity", name='cell_kernel')
            ]),
        OpTypePattern("Identity", name='cell_bias')
    ])]),
])

'''split pattern in lstm cell in tensorflow'''
lstm_xc_pattern = OpTypePattern('Split', inputs=[
    OpTypePattern("Const"), # axis for split
    OpTypePattern("BiasAdd", name="bias_add", inputs=[
        OpTypePattern("MatMul", inputs=[
            OpTypePattern("ConcatV2|Concat", name="xh"),
            OpTypePattern("*", name="cell_kernel"),
        ]),
        OpTypePattern("*", name="cell_bias"),
    ]),
])

''''gru cell pattern in tensorflow'''
grucell_pattern = \
    OpTypePattern('Add', name='gru_cell', inputs=[
        OpTypePattern('Mul', inputs=[
            gru_xc_pattern,
            OpTypePattern('*', name='input')
        ]),
        OpTypePattern('Mul', inputs=[
            OpTypePattern('Sub', inputs=[
                OpTypePattern('Const'),
                gru_xc_pattern
            ]),
            OpTypePattern('Tanh', inputs=[
                OpTypePattern('BiasAdd', inputs=[
                    OpTypePattern('MatMul', name='FullyConnect', inputs=[
                        OpTypePattern('Concat|ConcatV2', inputs=[
                            OpTypePattern('*', name='input'),
                            OpTypePattern('Mul', inputs=[
                                gru_xc_pattern,
                                OpTypePattern('*', name='input')
                            ]),
                            OpTypePattern('Const'),
                        ]),

                        OpTypePattern('Identity', name='candidate_kernel')
                    ]),
                    OpTypePattern('Identity', name='candidate_bias')
                ])
            ])
        ])
    ])


''''lstm cell pattern in tensorflow'''
lstmcell_pattern = \
    OpTypePattern('Mul', name='lstm_cell', inputs=[
        OpTypePattern("Sigmoid", name="ot", inputs=[lstm_xc_pattern]),
        OpTypePattern('Tanh', inputs=[
            OpTypePattern("Add", name="ct", inputs=[
                OpTypePattern("Mul", inputs=[
                    OpTypePattern("Sigmoid", name="ft", inputs=[
                        OpTypePattern("Add", inputs=[
                            lstm_xc_pattern,
                            OpTypePattern("*", name="ft_bias"),
                        ]),
                    ]),
                    OpTypePattern("*", name='input'),
                ]),
                OpTypePattern("Mul", inputs=[
                    OpTypePattern("Sigmoid", name="it", inputs=[lstm_xc_pattern]),
                    OpTypePattern("Tanh", name="gt", inputs=[lstm_xc_pattern]),
                ]),
            ]),
        ]),
    ])


rnn_patterns = {
    'tensorflow': {
        'gru_cell': grucell_pattern,
        'lstm_cell': lstmcell_pattern,
        'h_zero': static_rnn_h_zero_pattern,
        'static_rnn_batch_size': static_rnn_batch_size_pattern
    }
    # TODO: pytorch, mxnet, keras, cntk
}